""" Task object to generate / manage assessors and cluster."""

from datetime import date
import errno
import logging
import os
import shutil
import time

from . import cluster
from .cluster import PBS
from .errors import (NeedInputsException, NoDataException,
                     ClusterLaunchException)
from .dax_settings import DAX_Settings, DEFAULT_DATATYPE, DEFAULT_FS_DATATYPE
from . import assessor_utils


__copyright__ = 'Copyright 2013 Vanderbilt University. All Rights Reserved'
__all__ = ['Task', 'ClusterTask', 'XnatTask']
DAX_SETTINGS = DAX_Settings()
# Logger to print logs
LOGGER = logging.getLogger('dax')
# Job Statuses
# assessor that doesn't have data to run (for session assessor)
# E.G: dtiqa multi but no dti present.
NO_DATA = 'NO_DATA'
# assessor that is ready to be launch on the cluster.
# All the input data for the process to run are there.
NEED_TO_RUN = 'NEED_TO_RUN'
# assessor where input data are missing from a scan,
# multiple scans or other assessor.
NEED_INPUTS = 'NEED_INPUTS'
# the job has been submitted on the cluster and is running right now.
JOB_RUNNING = 'JOB_RUNNING'
# the job failed on the cluster.
JOB_FAILED = 'JOB_FAILED'
# Job done, waiting for the Spider to upload the results
READY_TO_UPLOAD = 'READY_TO_UPLOAD'
# in the process of uploading the resources on XNAT.
UPLOADING = 'UPLOADING'
# the assessors contains all the files. The upload and the job are done.
COMPLETE = 'COMPLETE'
# the job finished and upload is complete
READY_TO_COMPLETE = 'READY_TO_COMPLETE'
DOES_NOT_EXIST = 'DOES_NOT_EXIST'
OPEN_STATUS_LIST = [NEED_TO_RUN, UPLOADING, JOB_RUNNING, READY_TO_COMPLETE,
                    JOB_FAILED]
JOB_BUILT = 'JOB_BUILT'

# QC Statuses
# job is still running, not ready for QA yet
JOB_PENDING = 'Job Pending'
# job ready to be QA
NEEDS_QA = 'Needs QA'
# QC status set by the Image Analyst after looking at the results.
GOOD = 'Good'
# QC status set by the Image Analyst after looking at the results.
PASSED_QA = 'Passed'
# QC status set by the Image Analyst after looking at the results.
FAILED = 'Failed'
# QC status set by the Image Analyst after looking at the results.
BAD = 'Bad'
# QC status set by the Image Analyst after looking at the results.
POOR = 'Poor'
# will cause spider to delete results and rerun the processing
RERUN = 'Rerun'
# will cause spider to zip the current results and put in OLD,
# and then processing
REPROC = 'Reproc'
# Do not run this assessor anymore
DONOTRUN = 'Do Not Run'
FAILED_NEEDS_REPROC = 'Failed-needs reprocessing'  # FS
PASSED_EDITED_QA = 'Passed with edits'  # FS
OPEN_QA_LIST = [RERUN, REPROC]
BAD_QA_STATUS = [FAILED, BAD, POOR, DONOTRUN]

# Other Constants
DEFAULT_EMAIL_OPTS = DAX_SETTINGS.get_email_opts()
JOB_EXTENSION_FILE = DAX_SETTINGS.get_job_extension_file()
READY_TO_UPLOAD_FLAG_FILENAME = 'READY_TO_UPLOAD.txt'
OLD_RESOURCE = 'OLD'
EDITS_RESOURCE = 'EDITS'
REPROC_RES_SKIP_LIST = [OLD_RESOURCE, EDITS_RESOURCE]
INPUTS_DIRNAME = 'INPUTS'
BATCH_DIRNAME = 'BATCH'
OUTLOG_DIRNAME = 'OUTLOG'
PBS_DIRNAME = 'PBS'

# Status and QC status supported by DAX
SUPPORTED_STATUS = [NO_DATA, NEED_TO_RUN, NEED_INPUTS, JOB_RUNNING, JOB_FAILED,
                    READY_TO_UPLOAD, UPLOADING, READY_TO_COMPLETE, COMPLETE,
                    JOB_BUILT]
SUPPORTED_QC_STATUS = [JOB_PENDING, NEEDS_QA, GOOD, PASSED_QA, FAILED, BAD,
                       POOR, RERUN, REPROC, DONOTRUN, FAILED_NEEDS_REPROC,
                       PASSED_EDITED_QA]


def mkdirp(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            pass
        else:
            raise


def create_flag(flag_path):
    open(flag_path, 'w').close()


class Task(object):
    """ Class Task to generate/manage the assessor with the cluster """
    def __init__(self, processor, assessor, upload_dir):
        """
        Init of class Task

        :param processor: processor used
        :param assessor: pyxnat assessor object
        :param upload_dir: upload directory to copy data after job finished.
        :return: None

        """
        self.processor = processor
        self.assessor = assessor
        self.upload_dir = upload_dir
        self.atype = processor.xsitype.lower()

        # Cache for convenience
        self.assessor_label = assessor_utils.full_label_from_assessor(assessor)

    def get_processor_name(self):
        """
        Get the name of the Processor for the Task.

        :return: String of the Processor name.

        """
        return self.processor.name

    def get_processor_version(self):
        """
        Get the version of the Processor.

        :return: String of the Processor version.

        """
        return self.processor.version

    def is_open(self):
        """
        Check to see if a task is still in "Open" status as defined in
         OPEN_STATUS_LIST.

        :return: True if the Task is open. False if it is not open

        """
        astatus = self.get_status()
        return astatus in OPEN_STATUS_LIST

    def get_job_usage(self):
        """
        Get the amount of memory used, the amount of walltime used, the jobid
         of the process, the node the process ran on, and when it started
         from the scheduler.

        :return: List of strings. Memory used, walltime used, jobid, node used,
         and start date

        """
        atype = self.atype
        mgets = self.assessor.attrs.mget([
            '%s/memused' % atype,
            '%s/walltimeused' % atype,
            '%s/jobid' % atype,
            '%s/jobnode' % atype,
            '%s/jobstartdate' % atype
        ])
        return [mgets[0].strip(),
                mgets[1].strip(),
                mgets[2].strip(),
                mgets[3].strip(),
                mgets[4].strip()]

    def check_job_usage(self):
        """
        The task has now finished, get the amount of memory used, the amount of
         walltime used, the jobid of the process, the node the process ran on,
         and when it started from the scheduler. Set these values on XNAT

        :return: None

        """
        [memused, walltime, jobid, jobnode, jobstrdate] = self.get_job_usage()

        if walltime:
            if memused and jobnode:
                LOGGER.debug('memused and walltime already set, skipping')
            else:
                if memused == '':
                    self.set_memused('NotFound')
                if jobnode == '':
                    self.set_jobnode('NotFound')
            return

        # We can't get info from cluster if job too old
        if not cluster.is_traceable_date(jobstrdate):
            self.set_walltime('NotFound')
            self.set_memused('NotFound')
            self.set_jobnode('NotFound')
            return

        # Get usage with tracejob
        jobinfo = cluster.tracejob_info(jobid, jobstrdate)
        if jobinfo['mem_used'].strip():
            self.set_memused(jobinfo['mem_used'])
        else:
            self.set_memused('NotFound')
        if jobinfo['walltime_used'].strip():
            self.set_walltime(jobinfo['walltime_used'])
        else:
            self.set_walltime('NotFound')
        if jobinfo['jobnode'].strip():
            self.set_jobnode(jobinfo['jobnode'])
        else:
            self.set_jobnode('NotFound')

    def get_memused(self):
        """
        Get the amount of memory used for a process

        :return: String of how much memory was used

        """
        memused = self.assessor.attrs.get('%s/memused' % self.atype)
        return memused.strip()

    def set_memused(self, memused):
        """
        Set the amount of memory used for a process

        :param memused: String denoting the amount of memory used
        :return: None

        """
        self.assessor.attrs.set('%s/memused' % self.atype, memused)

    def get_walltime(self):
        """
        Get the amount of walltime used for a process

        :return: String of how much walltime was used for a process

        """
        walltime = self.assessor.attrs.get('%s/walltimeused' % self.atype)
        return walltime.strip()

    def set_walltime(self, walltime):
        """
        Set the value of walltime used for an assessor on XNAT

        :param walltime: String denoting how much time was used running
         the process.
        :return: None

        """
        self.assessor.attrs.set('%s/walltimeused' % self.atype, walltime)

    def get_jobnode(self):
        """
        Gets the node that a process ran on

        :return: String identifying the node that a job ran on

        """
        jobnode = self.assessor.attrs.get('%s/jobnode' % self.atype)
        if jobnode is None:
            jobnode = 'NotFound'
        return jobnode.strip()

    def set_jobnode(self, jobnode):
        """
        Set the value of the the node that the process ran on on the grid

        :param jobnode: String identifying the node the job ran on
        :return: None

        """
        self.assessor.attrs.set('%s/jobnode' % self.atype, jobnode)

    def undo_processing(self):
        """
        Unset the job ID, memory used, walltime, and jobnode information
         for the assessor on XNAT

        :except: pyxnat.core.errors.DatabaseError when attempting to
         delete a resource
        :return: None

        """
        from pyxnat.core.errors import DatabaseError

        self.set_qcstatus(JOB_PENDING)
        self.set_jobid(' ')
        self.set_memused(' ')
        self.set_walltime(' ')
        self.set_jobnode(' ')

        out_resource_list = self.assessor.out_resources()
        for out_resource in out_resource_list:
            if out_resource.label() not in REPROC_RES_SKIP_LIST:
                LOGGER.info('   Removing %s' % out_resource.label())
                try:
                    out_resource.delete()
                except DatabaseError:
                    LOGGER.error('   ERROR:deleting resource.')

    def reproc_processing(self):
        """
        If the procstatus of an assessor is REPROC on XNAT, rerun the assessor.

        :return: None

        """
        curtime = time.strftime("%Y%m%d-%H%M%S")
        local_dir = '%s_%s' % (self.assessor_label, curtime)
        local_zip = '%s.zip' % local_dir
        xml_filename = os.path.join(self.upload_dir, local_dir,
                                    '%s.xml' % self.assessor_label)

        # Make the temp dir
        mkdirp(os.path.join(self.upload_dir, local_dir))

        # Download the current resources
        out_resource_list = self.assessor.out_resources()
        for out_resource in out_resource_list:
            olabel = out_resource.label()
            if olabel not in REPROC_RES_SKIP_LIST and \
               len(out_resource.files().get()) > 0:
                LOGGER.info('   Downloading: %s' % olabel)
                out_res = self.assessor.out_resource(olabel)
                out_res.get(os.path.join(self.upload_dir, local_dir),
                            extract=True)

        # Download xml of assessor
        xml = self.assessor.get()
        with open(xml_filename, 'w') as f_xml:
            f_xml.write('%s\n' % xml)

        # Zip it all up
        cmd = 'cd %s && zip -qr %s %s/' % (self.upload_dir, local_zip,
                                           local_dir)
        LOGGER.debug('running cmd: %s' % cmd)
        os.system(cmd)

        # Upload it to Archive
        self.assessor.out_resource(OLD_RESOURCE)\
                     .file(local_zip)\
                     .put(os.path.join(self.upload_dir, local_zip))

        # Run undo
        self.undo_processing()

        # Delete the local copies
        os.remove(os.path.join(self.upload_dir, local_zip))
        shutil.rmtree(os.path.join(self.upload_dir, local_dir))

    def update_status(self):
        """
        Update the satus of a Task object.

        :return: the "new" status (updated) of the Task.

        """
        old_status, qcstatus, jobid = self.get_statuses()
        new_status = old_status

        if old_status == COMPLETE or old_status == JOB_FAILED:
            if qcstatus == REPROC:
                LOGGER.info('             * qcstatus=REPROC, running \
reproc_processing...')
                self.reproc_processing()
                new_status = NEED_TO_RUN
            elif qcstatus == RERUN:
                LOGGER.info('             * qcstatus=RERUN, running \
undo_processing...')
                self.undo_processing()
                new_status = NEED_TO_RUN
            else:
                # self.check_date()
                pass
        elif old_status == NEED_TO_RUN:
            # TODO: anything, not yet???
            pass
        elif old_status == READY_TO_COMPLETE:
            self.check_job_usage()
            new_status = COMPLETE
        elif old_status == NEED_INPUTS:
            # This is now handled by dax_build
            pass
        elif old_status == JOB_RUNNING:
            LOGGER.debug('calling check_running')
            new_status = self.check_running(jobid)
        elif old_status == READY_TO_UPLOAD:
            # TODO: let upload spider handle it???
            # self.check_date()
            pass
        elif old_status == UPLOADING:
            # TODO: can we see if it's really uploading???
            pass
        elif old_status == NO_DATA:
            pass
        else:
            LOGGER.warn('             * unknown status for %s: %s'
                        % (self.assessor_label, old_status))

        LOGGER.debug('new_status='+new_status)
        if new_status != old_status:
            LOGGER.info('             * changing status from %s to %s'
                        % (old_status, new_status))

            # Update QC Status
            if new_status == COMPLETE:
                self.set_proc_and_qc_status(new_status, NEEDS_QA)
            else:
                self.set_status(new_status)

        return new_status

    def get_jobid(self):
        """
        Get the jobid of an assessor as stored on XNAT

        :return: string of the jobid

        """
        jobid = self.assessor.attrs.get('%s/jobid' % self.atype)
        if jobid is None:
            jobid = 'NotFound'
        return jobid.strip()

    def get_job_status(self, jobid=None):
        """
        Get the status of a job given its jobid as assigned by the scheduler

        :param jobid: job id assigned by the scheduler
        :return: string from call to cluster.job_status or UNKNOWN.

        """
        jobstatus = 'UNKNOWN'
        if jobid is None:
            jobid = self.get_jobid()

        if jobid != '' and jobid != '0':
            jobstatus = cluster.job_status(jobid)

        LOGGER.debug('jobid,jobstatus='+str(jobid)+','+str(jobstatus))

        return jobstatus

    def launch(self, jobdir, job_email=None,
               job_email_options=DAX_SETTINGS.get_email_opts(),
               job_rungroup=None, xnat_host=None, writeonly=False, pbsdir=None,
               force_no_qsub=False):
        """
        Method to launch a job on the grid

        :param jobdir: absolute path where the data will be stored on the node
        :param job_email: who to email if the job fails
        :param job_email_options: grid-specific job email options (e.g.,
         fails, starts, exits etc)
        :param job_rungroup: grid-specific group to run the job under
        :param xnat_host: set the XNAT_HOST in the PBS job
        :param writeonly: write the job files without submitting them
        :param pbsdir: folder to store the pbs file
        :param force_no_qsub: run the job locally on the computer (serial mode)
        :raises: cluster.ClusterLaunchException if the jobid is 0 or empty
         as returned by pbs.submit() method
        :return: True if the job failed

        """
        cmds = self.commands(jobdir)
        pbsfile = self.pbs_path(writeonly, pbsdir)
        outlog = self.outlog_path()
        outlog_dir = os.path.dirname(outlog)
        mkdirp(outlog_dir)
        pbs = PBS(pbsfile,
                  outlog,
                  cmds,
                  self.processor.walltime_str,
                  self.processor.memreq_mb,
                  self.processor.ppn,
                  self.processor.env,
                  job_email,
                  job_email_options,
                  job_rungroup,
                  xnat_host,
                  self.processor.job_template)
        pbs.write()
        if writeonly:
            mes_format = """   filepath: {path}"""
            LOGGER.info(mes_format.format(path=pbsfile))
            return True
        else:
            jobid, job_failed = pbs.submit(outlog=outlog,
                                           force_no_qsub=force_no_qsub)

            if jobid == '' or jobid == '0':
                # TODO: check to be sure it didn't really launch, then
                # try one more time
                LOGGER.error('failed to launch job on cluster')
                raise ClusterLaunchException
            else:
                self.set_launch(jobid)
                if force_no_qsub or \
                   not cluster.command_found(DAX_SETTINGS.get_cmd_submit()):
                    if job_failed:
                        LOGGER.info('             * changing status to %s'
                                    % JOB_FAILED)
                        self.set_status(JOB_FAILED)
                    else:
                        LOGGER.info('             * changing status to %s'
                                    % READY_TO_UPLOAD)
                        # Status already set in the spider
                return True

    def check_date(self):
        """
        Sets the job created date if the assessor was not made through
         dax_build

        :return: Returns if get_createdate() is != '', sets date otherwise

        """
        if self.get_createdate() != '':
            return

        jobstartdate = self.get_jobstartdate()
        if jobstartdate != '':
            self.set_createdate(jobstartdate)

    def get_jobstartdate(self):
        """
        Get the date that the job started

        :return: String of the date that the job started in "%Y-%m-%d" format

        """
        return self.assessor.attrs.get('%s/jobstartdate' % self.atype)

    def set_jobstartdate_today(self):
        """
        Set the date that the job started on the grid to today

        :return: call to set_jobstartdate with today's date

        """
        today_str = str(date.today())
        return self.set_jobstartdate(today_str)

    def set_jobstartdate(self, date_str):
        """
        Set the date that the job started on the grid based on user passed
         value

        :param date_str: Datestring in the format "%Y-%m-%d" to set the job
         starte date to
        :return: None

        """
        self.assessor.attrs.set('%s/jobstartdate' % self.atype, date_str)

    def get_createdate(self):
        """
        Get the date an assessor was created

        :return: String of the date the assessor was created in "%Y-%m-%d"
         format

        """
        return self.assessor.attrs.get('%s/date' % self.atype)

    def set_createdate(self, date_str):
        """
        Set the date of the assessor creation to user passed value

        :param date_str: String of the date in "%Y-%m-%d" format
        :return: String of today's date in "%Y-%m-%d" format

        """
        self.assessor.attrs.set('%s/date' % self.atype, date_str)
        return date_str

    def set_createdate_today(self):
        """
        Set the date of the assessor creation to today

        :return: String of todays date in "%Y-%m-%d" format

        """
        today_str = str(date.today())
        self.set_createdate(today_str)
        return today_str

    def get_status(self):
        """
        Get the procstatus of an assessor

        :return: The string of the procstatus of the assessor.
         DOES_NOT_EXIST if the assessor does not exist

        """
        if not self.assessor.exists():
            xnat_status = DOES_NOT_EXIST
        elif self.atype.lower() in [DEFAULT_DATATYPE.lower(),
                                    DEFAULT_FS_DATATYPE.lower()]:
            xnat_status = self.assessor.attrs.get('%s/procstatus'
                                                  % self.atype.lower())
        else:
            xnat_status = 'UNKNOWN_xsiType: %s' % self.atype
        return xnat_status

    def get_statuses(self, cached_sessions=None):
        """
        Get the procstatus, qcstatus, and job id of an assessor

        :return: Serially ordered strings of the assessor procstatus,
         qcstatus, then jobid.
        """
        if cached_sessions:
            for csess in cached_sessions:
                for cassr in csess.assessors():
                    if cassr.label() == self.assessor_label:
                        pstatus = cassr.info()['procstatus']
                        qstatus = cassr.info()['qcstatus']
                        jobid = cassr.info()['jobid']
                        return pstatus, qstatus, jobid

        if not self.assessor.exists():
            xnat_status = DOES_NOT_EXIST
            qcstatus = DOES_NOT_EXIST
            jobid = ''
        elif self.atype.lower() in [DEFAULT_DATATYPE.lower(),
                                    DEFAULT_FS_DATATYPE.lower()]:
            xnat_status, qcstatus, jobid = self.assessor.attrs.mget([
                '%s/procstatus' % self.atype,
                '%s/validation/status' % self.atype,
                '%s/jobid' % self.atype
            ])
        else:
            xnat_status = 'UNKNOWN_xsiType: %s' % self.atype
            qcstatus = 'UNKNOWN_xsiType: %s' % self.atype
            jobid = ''

        return xnat_status, qcstatus, jobid

    def set_status(self, status):
        """
        Set the procstatus of an assessor on XNAT

        :param status: String to set the procstatus of the assessor to
        :return: None

        """
        self.assessor.attrs.set('%s/procstatus' % self.atype, status)

    def get_qcstatus(self):
        """
        Get the qcstatus of the assessor

        :return: A string of the qcstatus for the assessor if it exists.
         If it does not, it returns DOES_NOT_EXIST.
         The else case returns an UNKNOWN xsiType with the xsiType of the
         assessor as stored on XNAT.
        """
        qcstatus = ''

        if not self.assessor.exists():
            qcstatus = DOES_NOT_EXIST
        elif self.atype.lower() in [DEFAULT_DATATYPE.lower(),
                                    DEFAULT_FS_DATATYPE.lower()]:
            qcstatus = self.assessor.attrs.get('%s/validation/status'
                                               % self.atype)
        else:
            qcstatus = 'UNKNOWN_xsiType: %s' % self.atype

        return qcstatus

    def set_qcstatus(self, qcstatus):
        """
        Set the qcstatus of the assessor

        :param qcstatus: String to set the qcstatus to
        :return: None

        """
        self.assessor.attrs.mset({
            '%s/validation/status' % self.atype: qcstatus,
            '%s/validation/validated_by' % self.atype: 'NULL',
            '%s/validation/date' % self.atype: 'NULL',
            '%s/validation/notes' % self.atype: 'NULL',
            '%s/validation/method' % self.atype: 'NULL',
        })

    def set_proc_and_qc_status(self, procstatus, qcstatus):
        """
        Set the procstatus and qcstatus of the assessor

        :param procstatus: String to set the procstatus of the assessor to
        :param qcstatus: String to set the qcstatus of the assessor to
        :return: None

        """
        self.assessor.attrs.mset({
            '%s/procstatus' % self.atype: procstatus,
            '%s/validation/status' % self.atype: qcstatus,
        })

    def set_jobid(self, jobid):
        """
        Set the job ID of the assessor on XNAT

        :param jobid: The ID of the process assigned by the grid scheduler
        :return: None

        """
        self.assessor.attrs.set('%s/jobid' % self.atype, jobid)

    def set_launch(self, jobid):
        """
        Set the date that the job started and its associated ID on XNAT.
        Additionally, set the procstatus to JOB_RUNNING

        :param jobid: The ID of the process assigned by the grid scheduler
        :return: None

        """
        today_str = str(date.today())
        self.assessor.attrs.mset({
            '%s/jobstartdate' % self.atype.lower(): today_str,
            '%s/jobid' % self.atype.lower(): jobid,
            '%s/procstatus' % self.atype.lower(): JOB_RUNNING,
        })

    def commands(self, jobdir):
        """
        Call the get_cmds method of the class Processor.

        :param jobdir: Fully qualified path where the job will run on the node.
         Note that this is likely to start with /tmp on most grids.
        :return: A string that makes a command line call to a spider with all
         args.

        """
        assr_dir = os.path.join(jobdir, self.assessor_label)
        return self.processor.get_cmds(self.assessor, assr_dir)

    def pbs_path(self, writeonly=False, pbsdir=None):
        """
        Method to return the path of the PBS file for the job

        :param writeonly: write the job files without submitting them in TRASH
        :param pbsdir: folder to store the pbs file
        :return: A string that is the absolute path to the PBS file that will
         be submitted to the scheduler for execution.

        """
        res_dir = self.upload_dir
        j_ext = DAX_SETTINGS.get_job_extension_file()
        assessor_label = assessor_utils.full_label_from_assessor(
            self.assessor)
        filename = '%s%s' % (assessor_label, j_ext)
        if writeonly:
            if pbsdir and os.path.isdir(pbsdir):
                return os.path.join(pbsdir, filename)
            else:
                return os.path.join(os.path.join(res_dir, 'TRASH'), filename)
        else:
            if pbsdir:
                return os.path.join(pbsdir, filename)
            else:
                return os.path.join(os.path.join(res_dir, 'PBS'), filename)

    def outlog_path(self):
        """
        Method to return the path of outlog file for the job

        :return: A string that is the absolute path to the OUTLOG file.
        """
        assr_fout = '%s.output' % self.assessor_label
        return os.path.join(self.upload_dir, OUTLOG_DIRNAME, assr_fout)

    def ready_flag_exists(self):
        """
        Method to see if the flag file
        <UPLOAD_DIR>/<ASSESSOR_LABEL>/READY_TO_UPLOAD.txt exists

        :return: True if the file exists. False if the file does not exist.

        """
        flagfile = os.path.join(self.upload_dir, self.assessor_label,
                                READY_TO_UPLOAD_FLAG_FILENAME)
        return os.path.isfile(flagfile)

    def check_running(self, jobid=None):
        """
        Check to see if a job specified by the scheduler ID is still running

        :param jobid: The ID of the job in question assigned by the scheduler.
        :return: A String of JOB_RUNNING if the job is running or enqueued and
         JOB_FAILED if the ready flag (see read_flag_exists) does not exist
         in the assessor label folder in the upload directory.

        """
        # Check status on cluster
        jobstatus = self.get_job_status(jobid)

        LOGGER.debug('jobstatus='+str(jobstatus))

        if not jobstatus or jobstatus in ['R', 'Q']:
            # Still running
            return JOB_RUNNING
        elif not self.ready_flag_exists():
            # Check for a flag file created upon completion,
            # if it's not there then the job failed
            return JOB_FAILED
        else:
            # Let Upload Spider handle the upload
            return JOB_RUNNING


class ClusterTask(Task):
    """ Class Task to generate/manage the assessor with the cluster """
    def __init__(self, assr_label, upload_dir, diskq):
        """
        Init of class ClusterTask
        :return: None

        """
        self.assessor_label = assr_label
        self.processor = None
        self.assessor = None
        self.atype = None
        self.assessor_id = None
        self.diskq = diskq
        self.upload_dir = upload_dir

    def get_processor_name(self):
        """
        Get the name of the Processor for the Task.

        :return: String of the Processor name.

        """
        raise NotImplementedError()

    def get_processor_version(self):
        """
        Get the version of the Processor.

        :return: String of the Processor version.

        """
        raise NotImplementedError()

    def is_open(self):
        """
        Check to see if a task is still in "Open" status as defined in
         OPEN_STATUS_LIST.

        :return: True if the Task is open. False if it is not open

        """
        astatus = self.get_status()
        return astatus in OPEN_STATUS_LIST

    def get_job_usage(self):
        """
        Get the amount of memory used, the amount of walltime used, the jobid
         of the process, the node the process ran on, and when it started
         from the scheduler.

        :return: List of strings. Memory used, walltime used, jobid, node used,
                 and start date

        """
        memused = self.get_attr('memused')
        walltime = self.get_attr('walltimeused')
        jobid = self.get_attr('jobid')
        jobnode = self.get_attr('jobnode')
        jobstartdate = self.get_attr('jobstartdate')

        return [memused, walltime, jobid, jobnode, jobstartdate]

    def check_job_usage(self):
        """
        The task has now finished, get the amount of memory used, the amount of
         walltime used, the jobid of the process, the node the process ran on,
         and when it started from the scheduler. Set these values locally

        :return: None

        """
        [memused, walltime, jobid, jobnode, jobstrdate] = self.get_job_usage()

        if walltime:
            if memused and jobnode:
                LOGGER.debug('memused and walltime already set, skipping')
            else:
                if memused == '':
                    self.set_memused('NotFound')
                if jobnode == '':
                    self.set_jobnode('NotFound')
            return

        # We can't get info from cluster if job too old
        if not cluster.is_traceable_date(jobstrdate):
            self.set_walltime('NotFound')
            self.set_memused('NotFound')
            self.set_jobnode('NotFound')
            return

        # Get usage with tracejob
        jobinfo = cluster.tracejob_info(jobid, jobstrdate)
        if jobinfo['mem_used'].strip():
            self.set_memused(jobinfo['mem_used'])
        else:
            self.set_memused('NotFound')
        if jobinfo['walltime_used'].strip():
            self.set_walltime(jobinfo['walltime_used'])
        else:
            self.set_walltime('NotFound')
        if jobinfo['jobnode'].strip():
            self.set_jobnode(jobinfo['jobnode'])
        else:
            self.set_jobnode('NotFound')

    def get_memused(self):
        """
        Get the amount of memory used for a process

        :return: String of how much memory was used

        """
        memused = self.get_attr('memused')
        if memused is None:
            memused = 'NotFound'
        return memused

    def set_memused(self, memused):
        """
        Set the amount of memory used for a process

        :param memused: String denoting the amount of memory used
        :return: None

        """
        self.set_attr('memused', memused)

    def get_walltime(self):
        """
        Get the amount of walltime used for a process

        :return: String of how much walltime was used for a process

        """
        walltime = self.get_attr('walltimeused')
        if walltime is None:
            walltime = 'NotFound'

        return walltime

    def set_walltime(self, walltime):
        """
        Set the value of walltime used for an assessor

        :param walltime: String denoting how much time was used running
         the process.
        :return: None

        """
        self.set_attr('walltimeused', walltime)

    def get_jobnode(self):
        """
        Gets the node that a process ran on

        :return: String identifying the node that a job ran on

        """
        jobnode = self.get_attr('jobnode')
        return jobnode

    def set_jobnode(self, jobnode):
        """
        Set the value of the the node that the process ran on on the grid

        :param jobnode: String identifying the node the job ran on
        :return: None

        """
        self.set_attr('jobnode', jobnode)

    def undo_processing(self):
        raise NotImplementedError()

    def reproc_processing(self):
        """
        :raises: NotImplementedError
        :return: None

        """
        raise NotImplementedError()

    def update_status(self):
        """
        Update the status of a Cluster Task object.

        :return: the "new" status (updated) of the Task.

        """
        old_status = self.get_status()
        new_status = old_status
        LOGGER.debug('old_status='+old_status)

        if old_status == JOB_RUNNING:
            new_status = self.check_running()
            if new_status == READY_TO_UPLOAD:
                new_status = self.complete_task()
            elif new_status == JOB_FAILED:
                new_status = self.fail_task()
            else:
                # still running
                pass
        elif old_status in [COMPLETE, JOB_FAILED, NEED_TO_RUN,
                            NEED_INPUTS, READY_TO_UPLOAD, UPLOADING, NO_DATA]:
            pass
        else:
            LOGGER.warn('             * unknown status for %s: %s'
                        % (self.assessor_label, old_status))

        if new_status != old_status:
            LOGGER.info('             * changing status from %s to %s'
                        % (old_status, new_status))
            self.set_status(new_status)

        return new_status

    def get_jobid(self):
        """
        Get the jobid of an assessor as stored in local cache

        :return: string of the jobid

        """
        jobid = self.get_attr('jobid')
        return jobid

    def get_job_status(self):
        """
        Get the status of a job given its jobid as assigned by the scheduler

        :param jobid: job id assigned by the scheduler
        :return: string from call to cluster.job_status or UNKNOWN.

        """
        jobstatus = 'UNKNOWN'
        jobid = self.get_jobid()

        if jobid and jobid != '0':
            jobstatus = cluster.job_status(jobid)

        LOGGER.debug('jobid,jobstatus='+str(jobid)+','+str(jobstatus))

        return jobstatus

    def launch(self, force_no_qsub=False):
        """
        Method to launch a job on the grid

        :raises: cluster.ClusterLaunchException if the jobid is 0 or empty
         as returned by pbs.submit() method
        :return: True if the job failed

        """
        batch_path = self.batch_path()
        outlog = self.outlog_path()
        jobid, job_failed = cluster.submit_job(batch_path, outlog=outlog,
                                               force_no_qsub=force_no_qsub)

        if jobid == '' or jobid == '0':
            LOGGER.error('failed to launch job on cluster')
            raise ClusterLaunchException
        else:
            self.set_launch(jobid)
            cmd = DAX_SETTINGS.get_cmd_submit()
            if (force_no_qsub or not cluster.command_found(cmd)) and \
               job_failed:
                self.set_status(JOB_FAILED)
            return True

    def check_date(self):
        """
        Sets the job created date if the assessor was not made via dax_build
        """
        raise NotImplementedError()

    def get_jobstartdate(self):
        """
        Get the date that the job started

        :return: String of the date that the job started in "%Y-%m-%d" format

        """
        jobstartdate = self.get_attr('jobstartdate')
        if jobstartdate is None:
            jobstartdate = 'NULL'
        return jobstartdate

    def set_jobstartdate(self, date_str):
        """
        Set the date that the job started on the grid based on user passed
         value

        :param date_str: Datestring in the format "%Y-%m-%d" to set the job
         starte date to
        :return: None

        """
        self.set_attr('jobstartdate', date_str)

    def get_createdate(self):
        """
        Get the date an assessor was created

        :return: String of the date the assessor was created in "%Y-%m-%d"
         format

        """
        raise NotImplementedError()

    def set_createdate(self, date_str):
        """
        Set the date of the assessor creation to user passed value

        :param date_str: String of the date in "%Y-%m-%d" format
        :return: String of today's date in "%Y-%m-%d" format

        """
        raise NotImplementedError()

    def set_createdate_today(self):
        """
        Set the date of the assessor creation to today

        :return: String of todays date in "%Y-%m-%d" format

        """
        raise NotImplementedError()

    def get_status(self):
        """
        Get the procstatus

        :return: The string of the procstatus
        """
        procstatus = self.get_attr('procstatus')

        if not procstatus:
            procstatus = NEED_TO_RUN

        return procstatus

    def get_statuses(self):
        """
        Get the procstatus, qcstatus, and job id of an assessor
        """
        raise NotImplementedError()

    def set_status(self, status):
        """
        Set the procstatus of an assessor on XNAT

        :param status: String to set the procstatus of the assessor to
        :return: None

        """
        self.set_attr('procstatus', status)

    def get_qcstatus(self):
        """
        Get the qcstatus
        """
        raise NotImplementedError()

    def set_qcstatus(self, qcstatus):
        """
        Set the qcstatus of the assessor

        :param qcstatus: String to set the qcstatus to
        :return: None

        """
        raise NotImplementedError()

    def set_proc_and_qc_status(self, procstatus, qcstatus):
        """
        Set the procstatus and qcstatus of the assessor
        """
        raise NotImplementedError()

    def set_jobid(self, jobid):
        """
        Set the job ID of the assessor

        :param jobid: The ID of the process assigned by the grid scheduler
        :return: None

        """
        self.set_attr('jobid', jobid)

    def set_launch(self, jobid):
        """
        Set the date that the job started and its associated ID.
        Additionally, set the procstatus to JOB_RUNNING

        :param jobid: The ID of the process assigned by the grid scheduler
        :return: None

        """
        today_str = str(date.today())
        self.set_attr('jobstartdate', today_str)
        self.set_attr('jobid', jobid)
        self.set_attr('procstatus', JOB_RUNNING)

    def commands(self, jobdir):
        """
        Call the get_cmds method of the class Processor.

        :param jobdir: Fully qualified path where the job will run on the node.
         Note that this is likely to start with /tmp on most grids.
        :return: A string that makes a command line call to a spider with all
         args.

        """
        raise NotImplementedError()

    def batch_path(self):
        """
        Method to return the path of the PBS file for the job

        :return: A string that is the absolute path to the PBS file that will
         be submitted to the scheduler for execution.

        """
        label = self.assessor_label
        return os.path.join(self.diskq, BATCH_DIRNAME,
                            '%s%s' % (label, JOB_EXTENSION_FILE))

    def outlog_path(self):
        """
        Method to return the path of outlog file for the job

        :return: A string that is the absolute path to the OUTLOG file.
        """
        f_out = '%s.txt' % self.assessor_label
        return os.path.join(self.diskq, OUTLOG_DIRNAME, f_out)

    def processor_spec_path(self):
        """
        Method to return the path of processor file for the job

        :return: A string that is the absolute path to the file.
        """
        return os.path.join(self.diskq, 'processor', self.assessor_label)

    def upload_pbs_dir(self):
        """
        Method to return the path of dir for the PBS

        :return: A string that is the directory path for the PBS dir
        """
        label = self.assessor_label
        return os.path.join(self.upload_dir, label, PBS_DIRNAME)

    def upload_outlog_dir(self):
        """
        Method to return the path of outlog file for the job

        :return: A string that is the absolute path to the OUTLOG file.
        """
        label = self.assessor_label
        return os.path.join(self.upload_dir, label, OUTLOG_DIRNAME)

    def check_running(self):
        """
        Check to see if a job specified by the scheduler ID is still running

        :param jobid: The ID of the job in question assigned by the scheduler.
        :return: A String of JOB_RUNNING if the job is running or enqueued and
         JOB_FAILED if the ready flag (see read_flag_exists) does not exist
         in the assessor label folder in the upload directory.

        """

        if self.ready_flag_exists():
            return READY_TO_UPLOAD

        # Check status on cluster
        jobstatus = self.get_job_status()

        LOGGER.debug('jobstatus='+str(jobstatus))

        if not jobstatus or jobstatus == 'R' or jobstatus == 'Q':
            # Still running
            return JOB_RUNNING
        else:
            return JOB_FAILED

    def build_task(self):
        """
        Method to build a job
        """
        raise NotImplementedError()

    def build_commands(self):
        """
        Call the get_cmds method of the class Processor.

        :param jobdir: Fully qualified path where the job will run on the node.
         Note that this is likely to start with /tmp on most grids.
        :return: A string that makes a command line call to a spider with all
         args.

        """
        raise NotImplementedError()

    def get_attr(self, name):
        apath = self.attr_path(name)

        if not os.path.exists(apath):
            return None

        with open(apath, 'r') as f:
            return f.read().strip()

    def set_attr(self, name, value):
        attr_path = self.attr_path(name)
        attr_dir = os.path.dirname(attr_path)
        mkdirp(attr_dir)
        with open(self.attr_path(name), 'w') as f:
            f.write(str(value) + '\n')

    def attr_path(self, attr):
        return os.path.join(self.diskq, attr, self.assessor_label)

    def complete_task(self):
        self.check_job_usage()

        # Copy batch file, note we don't move so dax_upload knows the
        # task origin
        src = self.batch_path()
        dst = self.upload_pbs_dir()
        mkdirp(dst)
        LOGGER.debug('copying batch file from %s to %s' % (src, dst))
        shutil.copy(src, dst)

        # Move output file
        src = self.outlog_path()
        dst = self.upload_outlog_dir()
        if os.path.exists(src):
            mkdirp(dst)
            LOGGER.debug('moving outlog file from %s to %s' % (src, dst))
            shutil.move(src, dst)

        # Touch file for dax_upload to check
        res_dir = self.upload_dir
        create_flag(os.path.join(res_dir, self.assessor_label,
                    '%s.txt' % READY_TO_COMPLETE))

        return COMPLETE

    def fail_task(self):
        self.check_job_usage()

        # Copy batch file, note we don't move so dax_upload knows the
        # task origin
        src = self.batch_path()
        dst = self.upload_pbs_dir()
        mkdirp(dst)
        LOGGER.debug('copying batch file from %s to %s' % (src, dst))
        shutil.copy(src, dst)

        # Move output file
        src = self.outlog_path()
        dst = self.upload_outlog_dir()
        if os.path.exists(src):
            mkdirp(dst)
            LOGGER.debug('moving outlog file from %s to %s' % (src, dst))
            shutil.move(src, dst)

        # Touch file for dax_upload that job failed
        res_dir = self.upload_dir
        create_flag(os.path.join(res_dir, self.assessor_label,
                    '%s.txt' % JOB_FAILED))

        # Touch file for dax_upload to check
        create_flag(os.path.join(res_dir, self.assessor_label,
                    '%s.txt' % READY_TO_COMPLETE))

        return JOB_FAILED

    def delete_attr(self, attr):
        try:
            os.remove(self.attr_path(attr))
        except OSError:
            pass

    def delete_batch(self):
        # Delete batch file
        try:
            os.remove(self.batch_path())
        except OSError:
            pass

    def delete(self):
        # Delete attributes
        attr_list = ['jobid', 'jobnode', 'procstatus', 'walltimeused',
                     'memused', 'jobstartdate', 'processor']
        for attr in attr_list:
            self.delete_attr(attr)

        self.delete_batch()


class XnatTask(Task):
    """ Class Task to generate/manage the assessor with the cluster """
    def __init__(self, processor, assessor, upload_dir, diskq):
        """
        Init of class Task

        :param processor: processor used
        :param assessor: assessor dict ?
        :param upload_dir: upload directory to copy data when job finished.
        :return: None

        """
        super(XnatTask, self).__init__(processor, assessor, upload_dir)
        self.diskq = diskq

    def check_job_usage(self):
        """
        The task has now finished, get the amount of memory used, the amount of
         walltime used, the jobid of the process, the node the process ran on,
         and when it started from the scheduler. Set these values on XNAT

        :return: None

        """
        raise NotImplementedError()

    def update_status(self):
        """
        Update the satus of an XNAT Task object.

        :return: the "new" status (updated) of the Task.

        """
        old_status, qcstatus, jobid = self.get_statuses()
        new_status = old_status

        if old_status == COMPLETE or old_status == JOB_FAILED:
            if qcstatus == REPROC:
                LOGGER.info('             * qcstatus=REPROC, running \
reproc_processing...')
                self.reproc_processing()
                new_status = NEED_TO_RUN
            elif qcstatus == RERUN:
                LOGGER.info('             * qcstatus=RERUN, running \
undo_processing...')
                self.undo_processing()
                new_status = NEED_TO_RUN
            else:
                pass
        elif old_status in [NEED_TO_RUN, READY_TO_COMPLETE, NEED_INPUTS,
                            JOB_RUNNING, READY_TO_UPLOAD, UPLOADING, NO_DATA,
                            JOB_BUILT]:
            pass
        else:
            LOGGER.warn('             * unknown status for %s: %s'
                        % (self.assessor_label, old_status))

        if new_status != old_status:
            LOGGER.info('             * changing status from %s to %s'
                        % (old_status, new_status))
            self.set_status(new_status)

        return new_status

    def get_job_status(self):
        raise NotImplementedError()

    def launch(self):
        """
        Method to launch a job on the grid
        """
        raise NotImplementedError()

    def set_launch(self, jobid):
        """
        Set the date that the job started and its associated ID on XNAT.
        Additionally, set the procstatus to JOB_RUNNING

        :param jobid: The ID of the process assigned by the grid scheduler
        :return: None

        """
        raise NotImplementedError()

    def batch_path(self):
        """
        Method to return the path of the PBS file for the job

        :return: A string that is the absolute path to the PBS file that will
         be submitted to the scheduler for execution.

        """
        f_pbs = '%s%s' % (self.assessor_label, JOB_EXTENSION_FILE)
        return os.path.join(self.diskq, BATCH_DIRNAME, f_pbs)

    def outlog_path(self):
        """
        Method to return the path of outlog file for the job

        :return: A string that is the absolute path to the OUTLOG file.
        """
        f_txt = '%s.txt' % self.assessor_label
        return os.path.join(self.diskq, OUTLOG_DIRNAME, f_txt)

    def processor_spec_path(self):
        """
        Method to return the path of processor file for the job

        :return: A string that is the absolute path to the file.
        """
        return os.path.join(self.diskq, 'processor', self.assessor_label)

    def check_running(self):
        """
        Check to see if a job specified by the scheduler ID is still running

        :param jobid: The ID of the job in question assigned by the scheduler.
        :return: A String of JOB_RUNNING if the job is running or enqueued and
         JOB_FAILED if the ready flag (see read_flag_exists) does not exist
         in the assessor label folder in the upload directory.

        """
        raise NotImplementedError()

    def write_processor_spec(self):
        filename = self.processor_spec_path()
        mkdirp(os.path.dirname(filename))
        self.processor.write_processor_spec(filename)

    def build_task(self, assr, sessions,
                   jobdir, job_email=None,
                   job_email_options=DEFAULT_EMAIL_OPTS,
                   job_rungroup=None,
                   xnat_host=None):
        """
        Method to build a job
        """
        (old_proc_status, old_qc_status, _) = self.get_statuses(sessions)
        try:
            resdir = self.upload_dir
            cmds = self.build_commands(assr, sessions, jobdir, resdir)
            batch_file = self.batch_path()
            outlog = self.outlog_path()
            batch = PBS(batch_file,
                        outlog,
                        cmds,
                        self.processor.walltime_str,
                        self.processor.memreq_mb,
                        self.processor.ppn,
                        self.processor.env,
                        job_email,
                        job_email_options,
                        job_rungroup,
                        xnat_host,
                        self.processor.job_template)
            LOGGER.info('writing:' + batch_file)
            batch.write()

            # Set new statuses to be updated
            new_proc_status = JOB_RUNNING
            new_qc_status = JOB_PENDING

            # Write processor spec file for version 3
            try:
                if (self.processor.procyamlversion == '3.0.0-dev.0'):
                    # write processor spec file
                    LOGGER.debug('writing processor spec file')
                    self.write_processor_spec()
            except AttributeError as err:
                # older processor does not have version
                LOGGER.debug('procyamlversion not found'.format(err))

        except NeedInputsException as e:
            new_proc_status = NEED_INPUTS
            new_qc_status = e.value
        except NoDataException as e:
            new_proc_status = NO_DATA
            new_qc_status = e.value

        if new_proc_status != old_proc_status or \
           new_qc_status != old_qc_status:
            self.set_proc_and_qc_status(new_proc_status, new_qc_status)

        return (new_proc_status, new_qc_status)

    def build_commands(self, assr, sessions, jobdir, resdir):
        """
        Call the build_cmds method of the class Processor.

        :param jobdir: Fully qualified path where the job will run on the node.
         Note that this is likely to start with /tmp on most grids.
        :return: A string that makes a command line call to a spider with all
         args.

        """
        return self.processor.build_cmds(
            assr,
            self.assessor_label,
            sessions,
            jobdir,
            resdir)



from unittest import TestCase
from dax import utilities


class GroupbyToDictTest(TestCase):

    def test_find_with_pred(self):
        source = [{'a': 1, 'b': 2}, {'a': 2, 'b': 3}, {'a': 1, 'b': 4}]
        actual = utilities.find_with_pred(source, lambda x: x['a'] == 1)
        expected = {'a': 1, 'b': 2}
        self.assertDictEqual(actual, expected)

    def test_groupby_to_dict(self):
        source = [{'a': 1, 'b': 1}, {'a': 2, 'b': 2}, {'a': 3, 'b': 1}, {'a': 4, 'b': 2}]

        actual = utilities.groupby_to_dict(source, lambda x: x['b'])
        expected = {
            1: [{'a': 1, 'b': 1}, {'a': 3, 'b': 1}],
            2: [{'a': 2, 'b': 2}, {'a': 4, 'b': 2}]
        }

        self.assertDictEqual(actual, expected)

    def test_groupby_to_dict_2(self):
        source = [{'a': 1, 'b': 1}, {'a': 2, 'b': 2}, {'a': 3, 'b': 2}, {'a': 4, 'b': 1}]

        actual = utilities.groupby_to_dict(source, lambda x: x['b'])
        expected = {
            1: [{'a': 1, 'b': 1}, {'a': 4, 'b': 1}],
            2: [{'a': 2, 'b': 2}, {'a': 3, 'b': 2}]
        }

        self.assertDictEqual(actual, expected)

    def test_groupby_groupby_to_dict(self):
        source = [
            {'a': 1, 'b': 10, 'c': 100},
            {'a': 1, 'b': 11, 'c': 101},
            {'a': 1, 'b': 11, 'c': 102},
            {'a': 2, 'b': 12, 'c': 103},
            {'a': 1, 'b': 10, 'c': 104}
        ]
        actual = utilities.groupby_groupby_to_dict(source, lambda x: x['a'], lambda y: y['b'])
        expected = {
                1: {
                    10: [{'a': 1, 'b': 10, 'c': 100}, {'a': 1, 'b': 10, 'c': 104}],
                    11: [{'a': 1, 'b': 11, 'c': 101}, {'a': 1, 'b': 11, 'c': 102}]},
                2: {
                    12: [{'a': 2, 'b': 12, 'c': 103}]
                }
            }

        self.assertDictEqual(actual, expected)

    def test_strip_leading_and_trailing_spaces(self):
        tests = [
            ('', ''),
            (' ', ''),
            ('a', 'a'),
            (' a', 'a'),
            ('a ', 'a'),
            (' a ', 'a'),
            ('  a  ', 'a'),
            (' a , b ', 'a,b'),
        ]

        for t in tests:
            actual = utilities.strip_leading_and_trailing_spaces(t[0])
            self.assertEqual(actual, t[1])



import itertools as it
import json
import html
import fnmatch
import yaml
import os
import shutil
import re
import smtplib
from email.mime.text import MIMEText
import socket

from .errors import DaxError


def decode_url_json_string(json_string):
    """
    Load a string representing serialised json into
    :param json_string:
    :return:
    """
    strings = json.loads(html.unescape(json_string),
                         object_pairs_hook=parse_json_pairs)
    return strings


# TODO: BenM/assessor_of_assessor/document me!
# useful function for preventing key-value pairs in serialized json from being
# auto-converted to unicode
def parse_json_pairs(pairs):
    """
    An object hook for the json.loads method. Used in decode_url_json_string.
    :param pairs:
    :return: A dictionary of parsed json
    """
    sink_pairs = []
    for k, v in pairs:
        if isinstance(k, str):
            k = k.encode('utf-8').decode()
        if isinstance(v, str):
            v = v.encode('utf-8').decode()
        sink_pairs.append((k, v))
    return dict(sink_pairs)


def groupby_to_dict(source, group_pred):
    """
    Given a source iterable and a predicate defining how to group elements of
    the source iterable, convert the source iterable into a dictionary grouped
    by the key returned by the predicate.
    Example:
    source iterable: [{a:1, b:2}, (a:2, b:3}, {a:1, b:4}]
    group_pred: lambda x: x[a]

    results in:
    {
        1:[{a:1, b:2}, {a:1, b:4}],
        2:[{a:2, b:3}]
    }

    :param source: a keyless iterable (list or dictionary values or similar)
    :param group_pred: a function to determine the key by which each entry
    is grouped
    :return: the resulting dictionary of elements
    """
    results = dict()

    for k, v in it.groupby(source, group_pred):
        d = results.get(k, list())
        d.extend(list(v))
        results[k] = d

    return results


def groupby_groupby_to_dict(source, outer_pred, inner_pred):
    """
    Given a source iterable and two predicates defining how to group elements
    of the source iterable, convert the source iterable into a dictionary of
    dictionaries grouped by the keys returned by the predicate.

    :param source: a keyless iterable (list or dictionary values or similar)
    :param outer_pred: a function to determine the top level key by which each
    entry is grouped
    :param inner_pred: a function to determine the second level by which each
    entry is grouped
    :return: the resulting dictionary of dictionaries of elements
    """
    return {
        k: groupby_to_dict(v, inner_pred)
        for k, v in list(groupby_to_dict(source, outer_pred).items())
    }


def find_with_pred(items, pred):
    """
    Given a source iterable and a predicate defining how to identify an
    element, find and return the element if it is present, or None if the
    element is not found.

    :param items: a keyless iterable (list or dictionary values or similar)
    :param pred:
    :return: the element found by the predicate, or None
    """
    for i in items:
        if pred(i):
            return i
    return None


def strip_leading_and_trailing_spaces(list_arg):
    return ','.join([x.strip() for x in list_arg.split(',')])


def extract_exp(expression, full_regex=False):
    """Extract the experession with or without full_regex.

    :param expression: string to filter
    :param full_regex: using full regex
    :return: regex Object from re package
    """
    if not full_regex:
        exp = fnmatch.translate(expression)
    return re.compile(exp)


def clean_directory(directory):
    """
    Remove a directory tree or file

    :param directory: The directory (with sub directories if desired that you
     want to delete). Also works with a file.
    :return: None

    """
    for fname in os.listdir(directory):
        fpath = os.path.join(directory, fname)
        if os.path.isdir(fpath):
            shutil.rmtree(fpath)
        else:
            os.remove(fpath)


def check_image_format(fpath):
    """
    Check to see if a NIfTI file or REC file are uncompress and runs gzip via
     system command if not compressed

    :param fpath: Filepath of a NIfTI or REC file
    :return: the new file path of the gzipped file.

    """
    if fpath.endswith('.nii') or fpath.endswith('.rec'):
        os.system('gzip %s' % fpath)
        fpath = '%s.gz' % fpath
    return fpath


def read_yaml(yaml_file):
    """Functio to read a yaml file and return the document info

    :param yaml_file: yaml file path
    """
    with open(yaml_file, "r") as yaml_stream:
        try:
            return yaml.load(yaml_stream, Loader=yaml.FullLoader)
        except yaml.error.YAMLError as exc:
            err = 'YAML File {} could not be loaded properly. Error: {}'
            raise DaxError(err.format(yaml_file, exc))
    return None


def send_email(smtp_from, smtp_host, smtp_pass, to_addr, subject, content):
    """
    Send an email
    :param content: content of the email
    :param to_addr: address to send the email to
    :param subject: subject of the email
    :return: None
    """

    # Create the container (outer) email message.
    msg = MIMEText(content)
    msg['Subject'] = subject
    msg['From'] = smtp_from
    msg['To'] = ','.join(to_addr)

    # Send the email via our SMTP server
    smtp = smtplib.SMTP(smtp_host)
    smtp.starttls()
    smtp.login(smtp_from, smtp_pass)
    smtp.sendmail(smtp_from, to_addr, msg.as_string())
    smtp.quit()


def send_email_netrc(smtp_host, to_addr, subject, content):
    import netrc

    (smtp_from, _, smtp_pass) = netrc.netrc().authenticators(smtp_host)

    send_email(smtp_from, smtp_host, smtp_pass, to_addr, subject, content)


def get_this_instance():
    # build the instance name
    this_host = socket.gethostname().split('.')[0]
    this_user = os.environ['USER']
    return '{}@{}'.format(this_user, this_host)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""errors.py

Method related to errors and Custom Exceptions.
"""

import netrc
import sys


__copyright__ = 'Copyright 2013 Vanderbilt University. All Rights Reserved'
__all__ = ['DaxError', 'DaxXnatError', 'DaxProcessorError',
           'DaxSetupError', 'DaxNetrcError', 'DaxUploadError',
           'XnatAuthentificationError', 'XnatUtilsError', 'XnatAccessError',
           'XnatToolsError', 'XnatToolsUserError',
           'ClusterLaunchException', 'ClusterCountJobsException',
           'ClusterJobIDException',
           'AutoProcessorError']


# DAX error:
class DaxError(Exception):
    """Basic exception for errors raised by dax."""


# DAX XNAT error:
class DaxXnatError(DaxError):
    """Basic exception for errors related to XNAT raised by dax."""


# DAX Processor error:
class DaxProcessorError(DaxError):
    """Basic exception for errors related to processor raised by dax."""


# dax_setup errors
class DaxSetupError(DaxError, ValueError):
    """DaxSetup exception."""
    def __init__(self, message):
        Exception.__init__(self, 'Error in dax setup: %s' % message)


# Launcher errors:
class DaxUploadError(DaxError):
    """Custom exception raised with dax upload."""
    def __init__(self, message):
        Exception.__init__(self, 'Error with dax upload: %s' % message)


# Dax netrc errors
class DaxNetrcError(netrc.NetrcParseError):
    """Basic exception for errors related to DAX_Netrc raised by dax."""


# Launcher errors:
class DaxLauncherError(DaxError):
    """Custom exception raised with dax launcher."""
    def __init__(self, message):
        Exception.__init__(self, 'Error with Launcher: %s' % message)


# XnatUtils errors
class XnatAuthentificationError(DaxXnatError):
    """Custom exception raised when xnat connection failed."""
    def __init__(self, host=None, user=None):
        msg = 'ERROR: XNAT Authentification failed. Check logins.'
        if host:
            msg = '%s. Host: %s' % (msg, host)
            if user:
                msg = '%s / User: %s' % (msg, user)
        Exception.__init__(self, msg)


class XnatUtilsError(DaxXnatError):
    """XnatUtils exception."""
    def __init__(self, message):
        Exception.__init__(self, 'Error in XnatUtils: %s' % message)


class XnatToolsError(DaxError):
    """Xnat Tools Exception."""
    def __init__(self, message):
        Exception.__init__(self, 'Error in xnat_tools: %s' % message)


class XnatToolsUserError(DaxError):
    """Xnat Tools Exception."""
    def __init__(self, script, message):
        print(('\n%s: error: %s' % (script, message)))
        sys.exc_info()[-1]
        sys.exit(self)


class XnatAccessError(DaxXnatError, ValueError):
    """XNAT access exception if item does not exist."""
    def __init__(self, message):
        Exception.__init__(self, 'Error to access XNAT object: %s' % message)


# Cluster errors
class ClusterError(DaxError):
    """Basic exception for errors related to cluster raised by dax."""


class ClusterLaunchException(ClusterError):
    """Custom exception raised when launch on the grid failed"""
    def __init__(self):
        Exception.__init__(self, 'ERROR: Failed to launch job on the grid.')


class ClusterCountJobsException(ClusterError):
    """Custom exception raised when attempting to get the number of
    jobs fails"""
    def __init__(self):
        Exception.__init__(self, 'ERROR: Failed to fetch number of '
                                 'jobs from the grid.')


class ClusterJobIDException(ClusterError):
    """Custom exception raised when attempting to get the job id failed"""
    def __init__(self):
        Exception.__init__(self, 'ERROR: Failed to get job id.')


# Task:
class NeedInputsException(DaxError):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class NoDataException(DaxError):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


# Processor Exception:
class AutoProcessorError(DaxProcessorError):
    pass



from .suppdf import make_suppdf

if __name__ == '__main__':
    infile = '/Users/boydb1/TEST-fmri_msit/report_test3b.pdf'
    outfile = '/Users/boydb1/TEST-fmri_msit/report_test3b_sup.pdf'
    info = {}

    info['assessor'] = 'REMBRANDT-x-28034-x-28034a-x-fmri_msit_v2-x-6a8dcef0'
    info['proctype'] = 'fmri_msit_v2'
    info['procversion'] = '2.0.0'
    info['procdate'] = '2021/10/31'
    info['description'] = '''
1. Write conditions file
2. Write contrasts file
3. Realign/Motion-correction of FMRI
4. Coregister/Estimate mean FMRI to original T1
5. Segment & Normalize T1 to MNI space and segment
6. Apply warps and reslice FMRI and T1
7. ART outliers using CONN liberal settings
8. Skull-strip mean functional and apply to others
9. Create smoothed fmri'''

    info['inputs'] = (
        ('scan_anat', '3', '/projects/REMBRANDT/subjects/28043/experiments/28043a/scans/3'),
        ('scan_fmri', '14', '/projects/REMBRANDT/subjects/28043/experiments/28043a/scans/14'))
    
    info['outputs'] = [
        {
            'path': 'report*.pdf',
            'type': 'FILE',
            'resource': 'PDF'
        }, 
        {
            'path': 'stats.txt',
            'type': 'FILE',
            'resource': 'STATS',
        },
        {
            'path': 'PREPROC',
            'type': 'DIR',
            'resource': 'PREPROC',
        },
        {
            'path': '1stLEVEL',
            'type': 'DIR',
            'resource': '1stLEVEL'
        }
    ]

    info['session'] = {
        'PROJECT': 'REMBRANDT',
        'SUBJECT': '28043',
        'SESSION': '28043a'}

    info['proc'] = {
        'dax_version': '2.3.0',
        'dax_manager': 'vuiis_daily_singularity@hickory'}

    info['job'] = {
        'jobid': '33269491',
        'duration': '00:34:07',
        'memory': '2296336'}

    make_suppdf(infile, outfile, info)




from unittest import TestCase

from dax.processor_graph import ProcessorGraph
from dax import yaml_doc
from dax.tests import unit_test_common_processor_yamls as yamls
from dax.tests import unit_test_entity_common as common
from dax.processors import AutoProcessor


class TestLog:
    def __init__(self):
        self.warnings = list()

    def warning(self, message):
        self.warnings.append(message)

    def clear(self):
        self.warnings = list()


class ProcessorGraphUnitTests(TestCase):

    @staticmethod
    def __getabcdscenario():
        proc_a = yamls.generate_yaml(
            procname="Proc_A",
            scans=[{
            'name': 'scan1', 'types': 'T1',
            'resources': [
                {'type': 'NIFTI', 'name': 't1'}
            ]}
        ])
        proc_b = yamls.generate_yaml(
            procname="Proc_B",
            scans=[{
            'name': 'scan1', 'types': 'T1',
            'resources': [
                {'type': 'NIFTI', 'name': 't1'}
            ]
        }
        ])
        proc_c = yamls.generate_yaml(
            procname="Proc_C",
            assessors=[
            {
                'name': 'proc1', 'types': 'Proc_A_v1',
                'resources': [
                    {'type': 'SEG', 'name': 'proc_a'}
                ]
            },
            {
                'name': 'proc2', 'types': 'Proc_B_v1',
                'resources': [
                    {'type': 'SEG2', 'name': 'proc_b'}
                ]
            }
        ])
        proc_d = yamls.generate_yaml(
            procname="Proc_D",
            assessors=[
            {
                'name': 'proc1', 'types': 'Proc_C_v1',
                'resources': [
                    {'type': 'THING', 'name': 'proc_c'}
                ]
            },
            {
                'name': 'proc2', 'types': 'Proc_B_v1',
                'resources': [
                    {'type': 'SEG2', 'name': 'proc_b'}
                ]
            }
        ])

        return [
            ('Proc_A_v1', yaml_doc.YamlDoc().from_string(proc_a)),
            ('Proc_B_v1', yaml_doc.YamlDoc().from_string(proc_b)),
            ('Proc_C_v1', yaml_doc.YamlDoc().from_string(proc_c)),
            ('Proc_D_v1', yaml_doc.YamlDoc().from_string(proc_d))
        ]


    #def _getTestGraph1(self):



    def test_ordered_processors(self):
        log = TestLog()

        graph_description = {
            'a': [],
            'b': ['a'],
            'c': ['a'],
            'd': ['b'],
            'e': ['b', 'c'],
            'f': ['d', 'e']
        }
        actual = ProcessorGraph.order_from_inputs(graph_description, log)
        print(actual)


    def test_ordered_processors_has_cycle(self):
        log = TestLog()

        graph_description = {
            'a': [],
            'b': ['a', 'e'],
            'c': ['a'],
            'd': ['b'],
            'e': ['c', 'd'],
            'f': ['d', 'e']
        }
        actual = ProcessorGraph.order_from_inputs(graph_description, log)
        print(actual)

    def test_processor_inputs_from_sources(self):

        print((ProcessorGraph.processor_inputs_from_sources(
            ProcessorGraphUnitTests.__getabcdscenario()
        )))


    def test_ordering_from_sources(self):
        log = TestLog()
        print((ProcessorGraph.order_from_inputs(
            ProcessorGraph.processor_inputs_from_sources(
                ProcessorGraphUnitTests.__getabcdscenario()
            ),
            log
        )))

    def test_order_processors(self):
        yamldocs = [p[1] for p in ProcessorGraphUnitTests.__getabcdscenario()]
        processors = [AutoProcessor(common.FakeXnat, p) for p in yamldocs]
        log = TestLog()
        print((ProcessorGraph.order_processors(processors, log)))

    def test_order_processors_mocked(self):
        class TestProcessor:
            def __init__(self, name, inputs):
                self.name = name
                self.inputs = inputs

            def get_proctype(self):
                return self.name

            def get_assessor_input_types(self):
                return self.inputs

        log = TestLog()
        p = [
            TestProcessor('a', []),
            TestProcessor('b', []),
            TestProcessor('c', ['a', 'b']),
            TestProcessor(None, ['b']),
            TestProcessor('e', ['c', 'd']),
        ]

        actual = ProcessorGraph.order_processors(p, log)
        self.assertListEqual(
            actual, [p[0], p[1], p[2], p[4], p[3]]
        )
        self.assertListEqual(
            log.warnings,
            [
                'Unable to order all processors:',
                '  Unordered: e'
            ]
        )

        log = TestLog()
        p = [
            TestProcessor('a', []),
            TestProcessor('b', ['a', 'd']),
            TestProcessor('c', ['b']),
            TestProcessor('d', ['c']),
            TestProcessor('e', ['b'])
        ]

        actual = ProcessorGraph.order_processors(p, log)
        self.assertListEqual(
            actual,  [p[0], p[1], p[2], p[3], p[4]]
        )
        self.assertListEqual(
            log.warnings,
            [
                'Unable to order all processors:',
                '  Unordered: b, c, d, e',
                'Cyclic processor dependencies detected:',
                '  Cycle: d, c, b'
            ]
        )

        log = TestLog()
        p = [
            TestProcessor('a', []),
            TestProcessor('b', ['a', 'd']),
            TestProcessor('c', ['b']),
            TestProcessor('d', ['c']),
            TestProcessor('e', ['b', 'g']),
            TestProcessor('f', ['e']),
            TestProcessor('g', ['f']),
            TestProcessor('h', ['e'])
        ]

        actual = ProcessorGraph.order_processors(p, log)
        self.assertListEqual(
            actual,  [p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7]]
        )
        self.assertListEqual(
            log.warnings,
            [
                'Unable to order all processors:',
                '  Unordered: b, c, d, e, f, g, h',
                'Cyclic processor dependencies detected:',
                '  Cycle: d, c, b',
                '  Cycle: g, f, e'
            ]
        )


    def test_tarjan(self):

        def impl(g, expected):
            actual = ProcessorGraph.tarjan(g)
            print(actual)
            self.assertListEqual(actual, expected)

        g = {
            'a': ['b', 'c'],
            'b': ['d'],
            'c': ['e'],
            'd': ['f'],
            'e': ['f'],
            'f': []
        }
        impl(g, [['f'], ['d'], ['b'], ['e'], ['c'], ['a']])

        g = {
            'a': ['b'],
            'b': ['c', 'e'],
            'c': ['d'],
            'd': ['b'],
            'e': []
        }
        impl(g, [['e'], ['d', 'c', 'b'], ['a']])

        g = {
            'a': ['b'],
            'b': ['c'],
            'c': ['a']
        }
        impl(g, [['c', 'b', 'a']])

        g = {
            'a': ['b'],
            'b': ['c', 'e'],
            'c': ['d'],
            'd': ['b'],
            'e': ['f', 'h'],
            'f': ['g'],
            'g': ['e'],
            'h': []
        }
        impl(g, [['h'], ['g', 'f', 'e'], ['d', 'c', 'b'], ['a']])




import copy
import itertools
import logging
import sys
from collections import namedtuple

from .task import NeedInputsException
from .task import NEED_INPUTS, OPEN_STATUS_LIST, BAD_QA_STATUS,\
    JOB_PENDING, REPROC, RERUN, FAILED_NEEDS_REPROC, NEEDS_QA
from . import XnatUtils
from . import utilities

LOGGER = logging.getLogger('dax')

select_namespace = {
    'foreach': {'args': [{'optional': True, 'type': str}]},
    'one': {'args': []},
    'some': {'args': [{'optional': False, 'type': int}]},
    'all': {'args': []},
    'from': {'args': [{'optional': False, 'type': str}]}
}

select_session_namespace = {
    'current': {'args': []},
    'prior': {'args': [{'optional': False, 'type': int}]},
    'prior-with': {'args': [{'optional': False, 'type': int}]},
    'first': {'args': []},
    'first-with': {'args': []}
}


no_scans_error = 'No scan of the required type/s ({}) found for input {}'
no_asrs_error = 'No assessors of the require type/s ({}) found for input {}'
scan_unusable_error = 'Scan {} is unusable for input {}'
asr_unusable_error = 'Assessor {} is unusable for input {}'

missing_field_unnamed = 'Error: {} at position {} is missing {} field'
missing_field_named = "Error: {} '{}' is missing '{}' field"
bad_mode = ("Error: {} '{}': '{}' has an invalid value '{}'. "
            "It must be one of {}")
missing_resource_field_unnamed = \
    "Error in {} '{}': missing {} from resource at position {}"
missing_resource_field_named = \
    "Error in {} '{}': missing {} from resource field '{}'"
bad_resource_mode = \
    "Error in {} '{}'; resource field '{}' has an invalid value"

resource_paths = {
    'assessor': '{0}/out/resources/{1}',
    'scan': '{0}/resources/{1}'
}

uri_paths = {
    'assessor': '{0}/data{1}/out/resources/{2}',
    'scan': '{0}/data{1}/resources/{2}'
}


# parser pipeline
# . check whether artefacts of the appropriate type are present for a given
#   assessor
# . if they are, map them to inputs with the appropriate iteration
#   . if no foreach select statements are present, generate one set of command
#     parameters
#   . if one or more foreach select statements are present, generate the
#     appropriate cartesian product of command parameters
# . for each set of command parameters generated, create an assessor depending
#   on the state of the artefacts listed in the command parameters
#   . if one or more artefacts are of inappropriate quality


# BDB 6/5/21
# Removed inputs_by_type and related functions b/c it was not actually used.
# Also removed several class objects from ProcessorParser that 
# were only used within scope of parse_session function

class ParserArtefact:
    def __init__(self, path, resources, entity):
        self.name = path.split('/')[-1]
        self.path = path
        self.resources = resources,
        self.entity = entity

    def __repr__(self):
        return '{}(path = {}, resources = {}, entity = {})'.format(
            self.__class__.__name__, self.path, self.resources, self.entity
        )


class SelectSessionParameters:
    def __init__(self, mode, delta):
        self.mode = mode
        self.delta = delta

    def __repr__(self):
        return '{}(mode = {}, delta = {})'.format(self.__class__.__name__,
                                                  self.mode,
                                                  self.delta)


TimestampSession = namedtuple('TimestampSession', 'timestamp, session')

ArtefactEntry = namedtuple('ArtefactEntry', 'path, type, object')


class ProcessorParser:

    __schema_dict_v1 = {
        'top': set(['schema', 'inputs', 'xnat', 'attrs']),
        'xnat': set(['scans', 'assessors']),
        'scans': set(['select', 'types', 'nargs', 'resources', 'needs_qc']),
        'assessors': set(['select', 'types', 'nargs', 'resources', 'needs_qc']),
        'resources': set(['resource', 'varname', 'required'])
    }

    def __init__(self, yaml_source, proctype=None):
        (self.inputs, self.iteration_sources,
            self.iteration_map, self.prior_session_count) =\
            ProcessorParser.parse_inputs(yaml_source)

        self.match_filters = ProcessorParser.parse_match_filters(yaml_source)
        self.variables_to_inputs = ProcessorParser.parse_variables(self.inputs)
        self.xsitype = yaml_source['attrs'].get('xsitype', 'proc:genProcData')

        if proctype:
            self.proctype = proctype
        else:
            self.proctype = XnatUtils.get_proctype(
                yaml_source['inputs']['default']['spider_path'])[0]

        self.is_longitudinal_ = ProcessorParser.is_longitudinal(yaml_source)

    def parse_session(self, csess, sessions, pets=[]):
        """
        Parse a session to determine whether new assessors should be created.
        This call populates assessor_parameter_map.
        :param csess: the session in question
        :param sessions: the full list of sessions, including csess, for the
        subject
        :return: None
        """
        for i in range(len(sessions) - 1):
            if sessions[i].creation_timestamp() <\
                  sessions[i+1].creation_timestamp():
                raise ValueError("session param is not ordered by datetime")

        if not self.is_longitudinal_:
            relevant_sessions = [csess]
        else:
            index = sessions.index(csess)
            relevant_sessions = sessions[index:]

        # BDB 6/5/21
        # only include pets if this is the first mr session
        if sessions.index(csess) == (len(sessions) - 1):
            LOGGER.debug('session is first, including pets')
        else:
            LOGGER.debug('session is not first, not including pets')
            pets = []

        artefacts = ProcessorParser.parse_artefacts(relevant_sessions, pets)

        # BDB 6/5/21
        # The artefacts are a dictionary where the index key is the 
        # relative path of scan or assessor:
        # /projects/PROJ/subjects/SUBJ/experiments/SESS/assessors/ASSR
        # for every single assessor or scan. the value in the dictionary
        # is a ParserArtefact object the includes a list of the scan/assr's
        # resources and a CachedAssessor object. This can be used later
        # to quickly access this information

        # BDB 6/5/21
        # next we will create a dictionary of just the artefacts for each of
        # the inputs map the artefacts to the inputs, this is where
        # we filter down the whole session to the types of scan/assessors we
        # want. Then we decide what to do with the different combinations of
        # those scans/assessors if we find multiple per input. 
        # maybe we should change the names?
        # artefacts --> all_artefacts or all_session_arefacts
        # artefacts_by_inputs --> input_artefacts_by_input or something

        artefacts_by_input = ProcessorParser.map_artefacts_to_inputs(
                relevant_sessions, self.inputs, pets)

        # BDB 6/5/21
        # at this point the pet scan should be just like any other input or 
        # artefact, it's just a path

        # BDB 6/5/21
        # artefacts_by_input is a dictionary where the key is the
        # input name and the value is a list of artefact paths that match
        # the input.
        # These artefact paths are keys into the artefacts dictionary.

        parameter_matrix = \
            ProcessorParser.generate_parameter_matrix(
                self.inputs,
                self.iteration_sources,
                self.iteration_map,
                artefacts,
                artefacts_by_input)

        # BDB 6/5/21
        # parameter_matrix is the combinations of inputs from the lists in
        # artefacts_by_inputs. I think these are the cartesian product
        # of lists in artefacts_by_input.

        # BDB 6/5/21
        # Next we filter down the combinations by applying
        # any filters included in the yaml. currently
        # the only filter supported is a match filter
        # which help us only include combinations where one of the inputs
        # is the same, e.g. the same T1 input
        # This functions uses the artefacts dictionary to get the inputs field
        # from each artefact for comparison.

        parameter_matrix = ProcessorParser.filter_matrix(
            parameter_matrix,
            self.match_filters,
            artefacts)

        # BDB 6/5/21
        # And now we use the parameter matrix as a list of what set of inputs 
        # we need assessors for
        # by mapping to what assessors already exist by comparing
        # the inputs field on existing assessors with our list of inputs
        assessor_parameter_map = \
            ProcessorParser.compare_to_existing(relevant_sessions,
                                                self.proctype,
                                                parameter_matrix)

        # BDB 6/5/21
        # assessor_parameter_map is list of tuples
        # where each tuple is (inputs, assessor(s)) (if assesors exists already),
        # if assessors don't exist assessors will empty list

        # BDB 6/5/21
        # so what we are returning is a list of tuples
        # (set of inputs, existing asessors for these inputs)
        return list(assessor_parameter_map)

    def get_variable_set(self, assr):
        assr_inputs = XnatUtils.get_assessor_inputs(assr)

        # map from parameters to input resources
        command_set = dict()
        for k, v in list(self.variables_to_inputs.items()):
            inp = self.inputs[v['input']]
            artefact_type = inp['artefact_type']
            resource = v['resource']

            path_elements = [assr_inputs[v['input']], resource]

            command_set[k] =\
                resource_paths[artefact_type].format(*path_elements)

        return command_set

    def find_inputs(self, assr, sessions, assr_inputs):
        variable_set = {}
        input_list = []

        # Check artefact status
        LOGGER.debug('checking status of each artefact')
        for artk, artv in list(assr_inputs.items()):
            LOGGER.debug('checking status:' + artk)
            inp = self.inputs[artk]
            art_type = inp['artefact_type']

            if art_type == 'scan' and not inp['needs_qc']:
                continue

            if art_type == 'scan':
                # Check status of each input scan
                for vinput in artv:
                    qstatus = XnatUtils.get_scan_status(sessions, vinput)
                    if qstatus.lower() == 'unusable':
                        raise NeedInputsException(artk + ': Not Usable')
            else:
                # Check status of each input assr
                for vinput in artv:
                    pstatus, qstatus = XnatUtils.get_assr_status(sessions, vinput)
                    if pstatus in OPEN_STATUS_LIST + [NEED_INPUTS]:
                        raise NeedInputsException(artk + ': Not Ready')

                    if qstatus in [JOB_PENDING, REPROC, RERUN]:
                        raise NeedInputsException(artk + ': Not Ready')

                    if not inp['needs_qc']:
                        continue

                    if (qstatus in [FAILED_NEEDS_REPROC, NEEDS_QA]):
                        raise NeedInputsException(artk + ': Needs QC')

                    for badstatus in BAD_QA_STATUS:
                        if badstatus.lower() in qstatus.split(' ')[0].lower():
                            raise NeedInputsException(artk + ': Bad QC')

        # Map from parameters to input resources
        LOGGER.debug('mapping params to artefact resources')
        for k, v in list(self.variables_to_inputs.items()):
            LOGGER.debug('mapping:' + k)
            inp = self.inputs[v['input']]
            artefact_type = inp['artefact_type']
            resource = v['resource']

            # Find the resource
            cur_res = None
            for inp_res in inp['resources']:
                if inp_res['varname'] == k:
                    cur_res = inp_res
                    break

            # TODO: optimize this to get resource list only once
            for vnum, vinput in enumerate(assr_inputs[v['input']]):
                robj = assr._intf.select(
                    resource_paths[artefact_type].format(
                        vinput, resource))

                # Get list of all files in the resource, relative paths
                file_list = [x._urn for x in robj.files().get('path')]
                if len(file_list) == 0:
                    LOGGER.debug('empty or missing resource')
                    raise NeedInputsException('No Resource')

                if 'fmatch' in cur_res:
                    fmatch = cur_res['fmatch']
                elif cur_res['ftype'] == 'FILE':
                    # Default to all
                    fmatch = '*'
                else:
                    fmatch = None

                if 'filepath' in cur_res:
                    fpath = cur_res['filepath']
                    res_path = resource + '/files/' + fpath
                elif fmatch:
                    # Filter list based on regex matching
                    regex = utilities.extract_exp(fmatch, full_regex=False)
                    file_list = [x for x in file_list if regex.match(x)]

                    if len(file_list) == 0:
                        LOGGER.debug('no matching files found on resource')
                        raise NeedInputsException('No Files')

                    # Make a comma separated list of files
                    uri_list = ['{}/files/{}'.format(
                        resource, f) for f in file_list]
                    res_path = ','.join(uri_list)
                else:
                    res_path = resource + '/files'

                path_elements = [
                    assr._intf.host,
                    vinput,
                    res_path
                ]

                variable_set[k] = uri_paths[artefact_type].format(
                    *path_elements)

                if len(assr_inputs[v['input']]) > 1:
                    fdest = str(vnum) + cur_res['fdest']
                else:
                    fdest = cur_res['fdest']

                if 'ddest' in cur_res:
                    ddest = cur_res['ddest']
                else:
                    ddest = ''

                # Append to inputs to be downloaded
                input_list.append({
                    'fdest': fdest,
                    'ftype': cur_res['ftype'],
                    'fpath': variable_set[k],
                    'ddest': ddest
                })

                # Replace path with destination path after download
                if 'varname' in cur_res:
                    variable_set[k] = fdest

        LOGGER.debug('finished mapping params to artefact resources')

        return variable_set, input_list

    @staticmethod
    def _get_yaml_checker(version):
        if version == '1':
            return ProcessorParser.__check_yaml_v1
        return None

    @staticmethod
    def _get_schema_dictionary(version):
        if version == '1':
            return ProcessorParser.__schema_dict_v1

    @staticmethod
    def _check_valid_mode(input_category, input_name, keyword, valid_modes,
                          keyword_yaml):
        errors = []
        if keyword in keyword_yaml:
            mode = keyword_yaml[keyword]
            if mode not in valid_modes:
                valid_mode_str =\
                    ', '.join(["'" + x + "'" for x in valid_modes])
                errors.append(bad_mode.format(
                    input_category, input_name, keyword, mode, valid_mode_str))
        return errors

    @staticmethod
    def __check_resources_yaml_v1(input_category, input_name, resources_yaml):
        errors = []
        if 'resources' not in resources_yaml:
            errors.append(missing_field_named.format(
                input_category, input_name, 'resources'))

            for r, j in enumerate(resources_yaml['resources']):
                if 'varname' not in r:
                    errors.append(missing_resource_field_unnamed.format(
                        input_category, input_name, j, 'varname'))
                if 'resource' not in r:
                    errors.append(missing_resource_field_named.format(
                        input_category, input_name, r['varname'], 'resource'))
                if 'required' in r:
                    if r['required'] not in [True, False]:
                        errors.append(bad_resource_mode.format(
                            input_category, input_name, 'required'))
        return errors

    @staticmethod
    def __check_yaml_v1(log, yaml_source):

        # TODO: BenM/asr_of_asr/finish this!
        errors = []
        schema_number = yaml_source.get('yaml_processor_version', None)
        if schema_number != '0.1':
            errors.append('Error: Invalid schema num {}'.format(schema_number))

        if 'xnat' not in yaml_source:
            errors.append('Error: Missing xnat section')
        xnat_section = yaml_source['xnat']

        scan_section = xnat_section.get('scans', {})
        for s, i in enumerate(scan_section):
            if 'name' not in s:
                errors.append(missing_field_unnamed.format('scan', i, 'name'))
            name = s['name']

            if 'types' not in s:
                errors.append(
                    missing_field_named.format('scan', name, 'types'))

            errors.extend(
                ProcessorParser._check_valid_mode(
                    'scan', name, 'select', select_namespace, s))

            errors.extend(
                ProcessorParser._check_valid_mode(
                    'scan', name, 'select-session',
                    select_session_namespace, s))

            errors.extend(
                ProcessorParser.__check_resources_yaml_v1('scan', name, s))

        assr_section = xnat_section.get('assessors', {})
        for a, i in enumerate(assr_section):

            if 'name' not in a:
                errors.append(missing_field_unnamed.format('scan', i, 'name'))
            name = a['name']

            if 'types' not in a:
                errors.append(
                    missing_field_named.format('scan', name, 'types'))

            errors.extend(ProcessorParser._check_valid_mode(
                'assessor', name,
                'select', select_namespace, a))

            errors.extend(ProcessorParser._check_valid_mode(
                'assessor', name,
                'select-session', select_session_namespace, a))

            ProcessorParser.__check_resources_yaml_v1('assessor', name, a)

    @staticmethod
    def _get_args(statement):
        leftindex = statement.find('(')
        rightindex = statement.find(')')
        if leftindex == -1 and rightindex == -1:
            return [statement]
        elif leftindex != -1 and rightindex != -1:
            return [statement[:leftindex]] +\
                   [s.strip()
                    for s in statement[leftindex+1:rightindex].split(',')]
        else:
            raise ValueError('statement is malformed')

    @staticmethod
    def _parse_select(statement):
        if statement is None:
            statement = 'foreach'
        statement = statement.strip()
        return ProcessorParser._get_args(statement)

    @staticmethod
    def _parse_session_select(statement):
        if statement is None:
            statement = 'current'
        statement = statement.strip()
        args = ProcessorParser._get_args(statement)
        if args[0] == 'current':
            delta = 0
        elif args[0] in ['prior', 'prior-with']:
            delta = int(args[1])
        elif args[0] in ['first', 'first-with']:
            delta = sys.maxsize
        return SelectSessionParameters(args[0], delta)

    @staticmethod
    def _register_iteration_references(name, iteration_args, iteration_sources,
                                       iteration_map):
        if iteration_args[0] == 'foreach':
            if len(iteration_args) == 1:
                iteration_sources.add(name)
            else:
                iteration_map[name] = iteration_args[1]
        elif iteration_args[0] == 'one':
            iteration_sources.add(name)
        elif iteration_args[0] == 'all':
            iteration_sources.add(name)
        elif iteration_args[0] == 'from':
            iteration_map[name] = iteration_args[1]

    @staticmethod
    def _input_name(artefact):
        # candidates = list(filter(lambda v: v[1] is None, scan.iteritems()))
        # if len(candidates) != 1:
        #     raise ValueError(
        #         "invalid scan entry format; scan name cannot be determined")
        # return candidates[0][0]
        return artefact['name']

    @staticmethod
    def parse_inputs(yaml_source):
        # TODO: BenM/assessor_of_assessor/check error conditions on inputs:
        # . resource should be set
        # . no repeated input names
        # . ambiguous overlaps for input types (this may not be a problem)?

        iteration_sources = set()
        iteration_map = {}

        # get inputs: pass 1
        input_dict = yaml_source['inputs']
        xnat = input_dict['xnat']
        if xnat is None:
            raise ValueError(
                'yaml processor is missing xnat keyword contents')

        inputs = {}

        prior_session_count = 0

        # get scans
        scans = xnat.get('scans', list())
        for s in scans:
            name = ProcessorParser._input_name(s)

            select = s.get('select', None)
            parsed_select = ProcessorParser._parse_select(select)

            session_select = s.get('select-session', None)
            parsed_session_select =\
                ProcessorParser._parse_session_select(session_select)

            ProcessorParser._register_iteration_references(
                name,
                parsed_select,
                iteration_sources,
                iteration_map)

            types = [_.strip() for _ in s['types'].split(',')]
            resources = s.get('resources', [])
            artefact_required = False
            for r in resources:
                r['required'] = r.get('required', True)
                artefact_required = artefact_required or r['required']

            inputs[name] = {
                'types': types,
                'select': parsed_select,
                'select-session': parsed_session_select,
                'artefact_type': 'scan',
                'needs_qc': s.get('needs_qc', False),
                'resources': s.get('resources', []),
                'required': artefact_required
            }

            prior_session_count =\
                max(prior_session_count, parsed_session_select.delta)

        # get assessors
        asrs = xnat.get('assessors', list())
        for a in asrs:
            name = ProcessorParser._input_name(a)

            select = a.get('select', None)
            parsed_select = ProcessorParser._parse_select(select)

            session_select = a.get('select-session', None)
            parsed_session_select =\
                ProcessorParser._parse_session_select(session_select)

            ProcessorParser._register_iteration_references(
                name,
                parsed_select,
                iteration_sources,
                iteration_map)

            types = [_.strip() for _ in a['proctypes'].split(',')]
            resources = a.get('resources', [])
            artefact_required = False
            for r in resources:
                r['required'] = r.get('required', True)
            artefact_required = artefact_required or r['required']

            inputs[name] = {
                'types': types,
                'select': parsed_select,
                'select-session': parsed_session_select,
                'artefact_type': 'assessor',
                'needs_qc': a.get('needs_qc', False),
                'resources': a.get('resources', []),
                'required': artefact_required
            }

            prior_session_count =\
                max(prior_session_count, parsed_session_select.delta)

        # Handle petscans section
        petscans = xnat.get('petscans', list())
        for p in petscans:
            name = ProcessorParser._input_name(p)
            select = a.get('select', None)
            parsed_select = ProcessorParser._parse_select(select)
            parsed_session_select =\
                ProcessorParser._parse_session_select(session_select)
            ProcessorParser._register_iteration_references(
                name,
                parsed_select,
                iteration_sources,
                iteration_map)
            types = [x.strip() for x in p['scantypes'].split(',')]
            tracer = [x.strip() for x in p['tracer'].split(',')]
            inputs[name] = {
                'types': types,
                'select': parsed_select,
                'select-session': parsed_session_select,
                'artefact_type': 'scan',
                'needs_qc': p.get('needs_qc', False),
                'resources': p.get('resources', []),
                'required': True,
                'tracer': tracer}

        return (inputs, iteration_sources, iteration_map,
                prior_session_count)

    @staticmethod
    def is_longitudinal(yaml_source):
        inputs = yaml_source['inputs']['xnat']

        entries = inputs.get('scans', list()) + inputs.get('assessors', list())

        for e in entries:
            if e.get('select-session', 'current') != 'current':
                return True
        return False

    @staticmethod
    def parse_match_filters(yaml_source):
        match_list = []
        try:
            _filters = yaml_source['inputs']['xnat']['filters']
        except KeyError:
            return []

        # Parse out filters, currently only filters of type match are supported
        for f in _filters:
            _type = f['type']
            if _type == 'match':
                # Split the comma-separated list of inputs
                _inputs = f['inputs'].split(',')
                match_list.append(_inputs)
            else:
                LOGGER.error('invalid filter type:{}'.format(_type))

        return match_list

    @staticmethod
    def parse_variables(inputs):
        variables_to_inputs = {}
        for ik, iv in list(inputs.items()):
            for r in iv['resources']:
                v = r.get('varname', '')
                if v is not None and len(v) > 0:
                    variables_to_inputs[v] =\
                        {'input': ik, 'resource': r['resource']}

        return variables_to_inputs

    @staticmethod
    def parse_artefacts(csesses, pets=[]):
        def parse(carts, arts):
            for cart in carts:
                resources = {}
                for cres in cart.resources():
                    resources[cres.label()] = cres
                full_path = cart.full_path()
                arts[full_path] = ParserArtefact(full_path,
                                                 resources,
                                                 cart)

        artefacts = {}
        for csess in csesses:
            parse(csess.scans(), artefacts)
            parse(csess.assessors(), artefacts)

        # BDB 6/5/21
        # Add the pet scans (we are not supporting pet assessors at this time)
        for p in pets:
            parse(p.scans(), artefacts)

        return artefacts

    @staticmethod
    def map_artefacts_to_inputs(csesses, inputs, pets):

        # BDB 6/5/21
        # here is where we should do something different with for
        # the pet scans I think? are we treating assessors scans differently
        # here or not?

        artefacts_by_input = {k: [] for k in inputs}

        for i, iv in list(inputs.items()):
            # BDB 6/5/21 
            # here we do something to filter the list of sessions based
            # on the select types in the inputs???
            # I'm not sure what's going on here, are we only selecting
            # one of the sessions at this point? when and where
            # do we use multiple sessions?

            if iv['select-session'].mode in ['prior', 'prior-with']:
                if iv['select-session'].delta >= len(csesses):
                    csess = None
                else:
                    csess = csesses[iv['select-session'].delta]
            elif iv['select-session'].mode in ['first', 'first-with']:
                csess = csesses[-1]
            else:
                csess = csesses[0]

            # BDB 6/5/21 why would csess ever be None here??
            if csess is None:
                continue

            if 'tracer' in iv:
                # The input is a petscan so look in the pets
                for p in pets:
                    # Match the tracer name
                    tracer_name = p.get('xnat:tracer/name')
                    tracer_match = False
                    for expression in iv['tracer']:
                        regex = utilities.extract_exp(expression)
                        if regex.match(tracer_name):
                            # found a match so exit the loop
                            tracer_match = True
                            break

                    if not tracer_match:
                        # None of the expressions matched
                        LOGGER.debug('tracer no matchy:{}:{}'.format(
                            tracer_name, iv['tracer']))
                        continue
                    
                    # Now try to match the scan type
                    for pscan in p.scans():
                        for expression in iv['types']:
                            regex = utilities.extract_exp(expression)
                            if regex.match(pscan.type()):
                                # Found a match, now check quality
                                if pscan.info().get('quality') == 'unusable':
                                    LOGGER.info('excluding unusable scan')
                                else:
                                    artefacts_by_input[i].append(pscan.full_path())

            else:
                # Iterate each scan on the session
                for cscan in csess.scans():
                    for expression in iv['types']:
                        regex = utilities.extract_exp(expression)
                        if regex.match(cscan.type()):
                            if iv.get('select')[0] == 'all' and\
                                 cscan.info().get('quality') == 'unusable':
                                LOGGER.info('excluding unusable scan')
                            else:
                                artefacts_by_input[i].append(cscan.full_path())
                            # Break here so we don't match multiple times
                            break

                for cassr in csess.assessors():
                    try:
                        if cassr.type() in iv['types']:
                            artefacts_by_input[i].append(cassr.full_path())
                    except:
                        # Perhaps type/proctype is missing
                        LOGGER.error(f'Failed to add {cassr.full_label()} to processing list')
                        
        return artefacts_by_input

    # TODO: BenM improve name of generate_parameter_matrix
    # TODO: BenM handle multiple args disallowed / allowed scenarios
    @staticmethod
    def generate_parameter_matrix(inputs,
                                  iteration_sources,
                                  iteration_map,
                                  artefacts,
                                  artefacts_by_input):
        # generate n dimensional input matrix based on iteration sources
        all_inputs = []
        input_dimension_map = []

        # check whether all inputs are present
        for i, iv in list(inputs.items()):
            if len(artefacts_by_input[i]) == 0 and iv['required'] is True:
                return []

        # add in None for optional inputs so that the matrix can be generated
        # without artefacts present for those inputs
        sanitised_inputs = {}
        for i, iv in list(inputs.items()):
            if len(artefacts_by_input[i]) == 0:
                sanitised_inputs[i] = [list().append(None)]
            else:
                sanitised_inputs[i] = artefacts_by_input[i]

        for i in iteration_sources:
            # find other inputs that map to this iteration source
            mapped_inputs = [i]
            select_fn = inputs[i]['select'][0]

            # first, check iteration source and get the appropriate list of
            # artefacts
            cur_input_vector = None
            if select_fn == 'foreach':
                cur_input_vector = sanitised_inputs[i][:]

            elif select_fn == 'all':
                cur_input_vector = [[sanitised_inputs[i][:]]]

            elif select_fn == 'some':
                input_count = min(len(sanitised_inputs[i]),
                                  inputs[i]['select'][1])
                cur_input_vector = [[sanitised_inputs[i][:input_count]]]

            elif select_fn == 'one':
                cur_input_vector = [sanitised_inputs[i][:1]]

            if select_fn in ['foreach', 'from']:
                # build up the set of mapped input vectors one by one based on
                # the select mode of the mapped input
                combined_input_vector = [cur_input_vector]
                for k, v in list(iteration_map.items()):
                    if inputs[k]['select'][0] == 'foreach':
                        (v1, v2) = v, None
                    else:
                        (v1, v2) = v.split('/')
                    if v1 == i:
                        mapped_inputs.append(k)
                        if inputs[k]['select'][0] == 'foreach':
                            combined_input_vector.append(
                                sanitised_inputs[k][:])
                        else:  # from
                            from_artefacts = sanitised_inputs[v1]
                            mapped_input_vector = []
                            for fa in from_artefacts:
                                a = artefacts[fa]
                                from_inputs = a.entity.get_inputs()
                                if from_inputs is not None:
                                    mapped_input_vector.append(
                                        from_inputs[v2])

                            combined_input_vector.append(mapped_input_vector)

                    else:
                        pass

                # 'trim' the input vectors to the number of entries of the
                # shortest vector. We don't actually truncate the datasets but
                # just use the number when transposing, below
                min_entry_count = min((len(e) for e in combined_input_vector))

                # transpose from list of input vectors to input entry lists,
                # one per combination of inputs
                merged_input_vector = [
                    [None for col in range(len(combined_input_vector))]
                    for row in range(min_entry_count)]
                for row in range(min_entry_count):
                    for col in range(len(combined_input_vector)):
                        merged_input_vector[row][col] =\
                            combined_input_vector[col][row]

                all_inputs.append(mapped_inputs)
                input_dimension_map.append(merged_input_vector)

            else:
                all_inputs.append(mapped_inputs)
                input_dimension_map.append(cur_input_vector)

        # perform a cartesian product of the dimension map entries to get the
        # final input combinations
        matrix = [list(
            itertools.chain.from_iterable(x)) for x in itertools.product(
                *input_dimension_map)]

        matrix_headers = list(itertools.chain.from_iterable(all_inputs))

        # rebuild the matrix to order the inputs consistently
        final_matrix = []
        for r in matrix:
            row = dict()
            for i in range(len(matrix_headers)):
                row[matrix_headers[i]] = r[i]
            final_matrix.append(row)

        return final_matrix

    @staticmethod
    def compare_to_existing(csesses, proc_type, parameter_matrix):

        csess = csesses[0]

        assessors = [[] for _ in range(len(parameter_matrix))]

        existing_assessors = []

        # Get of list assessors that have already been
        # created on xnat for this processor/proctype
        for casr in csess.assessors():
            try:
                proc_type_matches = (casr.type() == proc_type)
                if proc_type_matches:
                    existing_assessors.append(casr)
            except:
                LOGGER.error(f'Failed to check type of {casr.label()}')
                continue

        # Check for any missing inputs fields in the existing assessors
        for casr in existing_assessors:
            # Check for empty inputs. If ANY of the assessors on this session
            # have an empty inputs, then we cannot determine which input sets
            # need to be built. We refuse to do anything by returning
            # an empty list.
            if casr.get_inputs() is None:
                LOGGER.warn('assessor with empty inputs field, cannot build processor for session:' + casr.label())
                return list()

        # Compare existing to the "to build" assessors in parameter_matrix
        for casr in existing_assessors:
            for pi, p in enumerate(parameter_matrix):
                if casr.get_inputs() == p:
                    # BDB  6/5/21 do we ever have more than one assessor
                    #             with the same set of inputs?
                    assessors[pi].append(casr)

        return list(zip(copy.deepcopy(parameter_matrix), assessors))

    @staticmethod
    def get_input_value(input_name, parameter, artefacts):
        if '/' not in input_name:
            # Matching on parent so keep this value
            _val = parameter[input_name]
        else:
            # Match is on a parent so parse out the parent/child
            (_parent_name, _child_name) = input_name.split('/')
            _parent_val = parameter[_parent_name]
            _parent_art = artefacts[_parent_val]

            _parent_art_inputs = _parent_art.entity.get_inputs()
            if _parent_art_inputs is None:
                # Check that inputs field is not empty
                LOGGER.warn('inputs field is empty:' + _parent_val)
                _val = None
            else:
                # Get the inputs field from the child
                _parent_inputs = _parent_art_inputs
                _val = _parent_inputs[_child_name]

        return _val

    @staticmethod
    def filter_matrix(parameter_matrix, match_filters, artefacts):
        filtered_matrix = []
        for cur_param in parameter_matrix:
            # Reset matching for this param set
            all_match = True

            for cur_filter in match_filters:
                # Get the first value to compare with others
                first_val = ProcessorParser.get_input_value(
                    cur_filter[0], cur_param, artefacts)

                # Compare other values with first value
                for cur_input in cur_filter[1:]:
                    cur_val = ProcessorParser.get_input_value(
                        cur_input, cur_param, artefacts)

                    if cur_val is None:
                        LOGGER.warn('cannot match, empty inputs:{}'.format(
                            cur_input))
                        all_match = False
                        break

                    if cur_val != first_val:
                        # A single non-match breaks the whole thing
                        all_match = False
                        break

            if all_match:
                # Keep this param set if everything matches
                filtered_matrix.append(cur_param)

        return filtered_matrix



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Created on 16th February, 2017:

File containing all useful functions for the different executables.

@author: Benjamin Yvernault, Electrical Engineering, Vanderbilt University
'''

import logging
import os
import sys

from .errors import XnatToolsError, XnatToolsUserError

LENGTH = 64
DISPLAY_TEMPLATE = """#######################################################\
#########
{name}
#                                                              #
# Developed by the MASI Lab Vanderbilt University, TN, USA.    #
# If issues, please start a thread here:                       #
# https://groups.google.com/forum/#!forum/vuiis-cci            #
# Usage:                                                       #
{description}
# Examples:                                                    #
#     Check the help for examples by running --help            #
################################################################
{extra}"""
ARGS_DISPLAY = """Arguments:
{args}"""
CSV_HEADER = [
    'object_type', 'project_id', 'subject_label', 'session_type',
    'session_label', 'as_label', 'as_type', 'as_description', 'as_quality',
    'resource', 'fpath'
]
ORDER = ['commun', 'project', 'subject', 'session', 'scan', 'assessor']
XNAT_MODALITIES = {
    'CT': {'xsitype': 'xnat:ctSessionData',
           'info': 'An event in which CT scans are obtained on a subject'},
    'MR': {'xsitype': 'xnat:mrSessionData',
           'info': 'An event in which MR scans are obtained on a subject'},
    'PET': {'xsitype': 'xnat:petSessionData',
            'info': 'An event in which PET scans are obtained on a subject'},
    'EPS': {'xsitype': 'xnat:epsSessionData',
            'info': 'Cardiac Electrophysiology Session'},
    'DX': {'xsitype': 'xnat:dxSessionData',
           'info': 'An event in which Digital Radiography scans are obtained \
on a subject'},
    'RT': {'xsitype': 'xnat:rtSessionData', 'info': 'Radiotherapy Session'},
    'EEG': {'xsitype': 'xnat:eegSessionData',
            'info': 'Electroencephalography Session'},
    'HD': {'xsitype': 'xnat:hdSessionData', 'info': 'Hemodynamic Session'},
    'DX3DCRANIOFACIAL': {'xsitype': 'xnat:dx3DCraniofacialSessionData',
                         'info': 'X-Ray 3D Craniofacial Session'},
    'ECG': {'xsitype': 'xnat:ecgSessionData',
            'info': 'Electrocardiography Session'},
    'OTHERDICOM': {'xsitype': 'xnat:otherDicomSessionData',
                   'info': 'DICOM study of undetermined type'},
    'RF': {'xsitype': 'xnat:rfSessionData',
           'info': 'Radiofluoroscopy Session'},
    'XA3D': {'xsitype': 'xnat:xa3DSessionData',
             'info': 'X-Ray 3D Angiography Session'},
    'ESV': {'xsitype': 'xnat:esvSessionData',
            'info': 'Video Endoscopy Session'},
    'XC': {'xsitype': 'xnat:xcSessionData',
           'info': 'Visible Light Photography Session'},
    'XA': {'xsitype': 'xnat:xaSessionData',
           'info': 'An event in which X-ray Angiography scans are obtained on \
a subject'},
    'MEG': {'xsitype': 'xnat:megSessionData',
            'info': 'Magnetoencephalography Session'},
    'IO': {'xsitype': 'xnat:ioSessionData',
           'info': 'Intraoral Radiography Session'},
    'CR': {'xsitype': 'xnat:crSessionData',
           'info': 'Computed Radiography Session'},
    'GM': {'xsitype': 'xnat:gmSessionData',
           'info': 'Visible Light Microscopy Session'},
    'GMV': {'xsitype': 'xnat:gmvSessionData',
            'info': 'Video Microscopy Session'},
    'ES': {'xsitype': 'xnat:esSessionData',
           'info': 'Visible Light Endoscopy Session'},
    'OPT': {'xsitype': 'xnat:optSessionData',
            'info': 'Ophthalmic Tomography Session'},
    'MG': {'xsitype': 'xnat:mgSessionData',
           'info': 'Digital Mammography Session'},
    'US': {'xsitype': 'xnat:usSessionData',
           'info': 'Ultrasound Session'},
    'XCV': {'xsitype': 'xnat:xcvSessionData',
            'info': 'Video Photography Session'},
    'SM': {'xsitype': 'xnat:smSessionData',
           'info': 'Visible Light Slide-Coordinates Microscopy Session'},
    'OP': {'xsitype': 'xnat:opSessionData',
           'info': 'Ophthalmic Photography Session'},
}


def parse_args(name, description, add_tools_arguments, purpose,
               extra_display=''):
    """
    Method to parse arguments base on argparse

    :param name: name of the script
    :param description: description of the script for help display
    :param add_tools_arguments: fct to add arguments to parser
    :param purpose: purpose of the script
    :param extra_display: extra display
    :return: parser object
    """
    from argparse import ArgumentParser, RawTextHelpFormatter
    parser = ArgumentParser(prog=name, description=description,
                            formatter_class=RawTextHelpFormatter)
    parser.add_argument('--host', dest='host', default=None,
                        help='Host for XNAT. Default: env XNAT_HOST.')
    parser.add_argument('-u', '--username', dest='username', default=None,
                        help='Username for XNAT.')
    parser = add_tools_arguments(parser)
    main_display(name, purpose, extra_display)
    args = parser.parse_args()
    args_display(args)
    return args


def main_display(name, description, extra_display=''):
    """
    Main display of the executables before any process

    :param name: name of the executable
    :param description: description to display
    :param extra_display: extra display
    :return: None
    """
    _spaces = (LENGTH - len(name)) // 2
    _name = edit_string_size(name, left_spaces=_spaces)
    _desc = edit_string_size(description, left_spaces=4)
    print((DISPLAY_TEMPLATE.format(name=_name, description=_desc,
                                  extra=extra_display)))
    print_separators(symbol='-')


def args_display(args):
    """
    Display arguments set by user

    :param args: arguments from parse_args
    :param description: description to display
    :return: None
    """
    _args = list()
    for key, value in list(vars(args).items()):
        _format = '  %*s -> %*s'
        if isinstance(value, bool):
            if value:
                _args.append(_format % (-15, key.replace('_', ' '), -43, 'on'))
        else:
            if value:
                _args.append(_format % (-15, key.replace('_', ' '),
                                        -43, get_proper_str(value, True)))

    print((ARGS_DISPLAY.format(args='\n'.join(_args))))
    print_separators(symbol='-', return_line=True)


def run_tool(script, description, add_tools_arguments, purpose, run_tool_fct,
             extra_display=''):
    """
    Main function to run for all xnat tools.
    Set args/display and run core function (run_tool_fct)
    See tools for example.

    :param script: name of script
    :param add_tools_arguments: fct adding arguments to parser
    :param display: dictionary coding display
    :param run_tool_fct: core fct for the tool
    :param extra_display: extra display after checking args
    :return: None
    """
    args = parse_args(script, description, add_tools_arguments, purpose,
                      extra_display)
    run_tool_fct(args)


def setup_info_logger(name, log_file=None):
    """
    Using logger for the executables output.
     Setting the information for the logger.

    :param name: Name of the logger
    :param log_file: log file path to write outputs
    :return: logging object
    """
    if log_file:
        handler = logging.FileHandler(log_file, 'w')
    else:
        handler = logging.StreamHandler(sys.stdout)

    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    logger.addHandler(handler)
    return logger


def edit_string_size(strings, max_length=LENGTH - 4, left_spaces=0,
                     sformat='# %s%s #', symbol=' '):
    """
    Edit the string to by adding spaces at the beginning and end.
    If string is a list, return list of strings edited.

    :param string: string to edit or list of strings
    :param length: length of the string
    :return: new string of length 60
    """
    if isinstance(strings, str):
        _lspaces = symbol * int(left_spaces)
        if symbol != ' ':
            _lspaces = '%s ' % _lspaces[:-1]
        _str = '%s%s' % (_lspaces, strings)
        _space = (max_length - len(_str))
        return sformat % (_str, symbol * int(_space))
    elif isinstance(strings, list):
        # Separate the string in several
        new_strings = list()
        for string in strings:
            _str = edit_string_size(string, max_length, left_spaces)
            new_strings.append(_str)
        return '\n'.join(new_strings)
    else:
        err = "Wrong type for 'strings' argument in edit_string_size. Required\
 <type 'str'> or <type 'list'>, found: %s"
        raise XnatToolsError(err % type(string))


def read_txt(txt_file, exe_name=''):
    """
    Method to read the txt file path with per line, the name
     of the variable on REDCap you want to extract

    :param txt_file: filepath
    :return: list of REDCap variables
    """
    if txt_file:
        print(('INFO: Export data from text file %s ...' % txt_file))
        obj_list = list()
        if not os.path.exists(txt_file):
            err = 'file %s does not exist.'
            raise XnatToolsUserError(exe_name, err % txt_file)
        else:
            with open(txt_file, 'r') as input_file:
                for line in input_file:
                    obj_list.append(line.strip().split('\n')[0])
    else:
        obj_list = None
    return obj_list


def write_csv(csv_string, csv_file, exe_name=''):
    """
    Method to write the report as a csv file
     with the values from REDCap

    :param csv_string: data to write in the csv
    :param csv_file: csv filepath
    :param exe_name: name of executable running the function for error
    :return: None
    """
    print('INFO: Writing report ...')
    basedir = os.path.basedir(csv_file)
    if not os.path.exists(basedir):
        err = 'Path %s not found for report. Give an existing parent folder.'
        raise XnatToolsUserError(exe_name, err % csv_file)
    with open(csv_file, 'w') as output_file:
        for line in csv_string:
            output_file.write(line)


def get_option_list(string, all_value='all'):
    """
    Method to switch the string of value separated by a comma into a list.
     If the value is 'all', keep all.

    :param string: string to change
    :param all_value: value to return if all used
    :return: None if not set, all if all, list of value otherwise
    """
    if not string or string == 'nan':
        return None
    elif string == 'all':
        return all_value
    else:
        return string.split(',')


def get_proper_str(str_option, end=False, size=43):
    """
    Method to shorten a string into the proper size for display

    :param str_option: string to shorten
    :param end: keep the end of the string visible (default beginning)
    :return: shortened string
    """
    if len(str_option) > size:
        if end:
            return '...%s' % str_option[-(size - 3):]
        else:
            return '%s...' % str_option[:(size - 3)]
    else:
        return str_option


def prompt_user_yes_no(question):
    """Prompt the user for a question with answer Y/N.

    :return: True if yes, False if no, ask again if any other answer
    """
    value = ''
    while value.lower() not in ['yes', 'no', 'n', 'y']:
        value = input("%s [yes/no] " % question)

    if value.lower() in ['yes', 'y']:
        return True
    else:
        return False


def print_separators(symbol='=', length=LENGTH, return_line=False):
    """
    Print line to separate: symbolx60

    :param symbol: symbol to print length time (60)
    """
    _format = '%s'
    if return_line:
        _format = '%s\n'
    print((_format % (symbol * length)))


def print_end(name):
    """
    Last display when the tool script finish.
    """
    _spaces = (LENGTH - 6 - len(name)) / 2
    print(('%s\n' % edit_string_size(name, max_length=LENGTH - 6,
                                    left_spaces=_spaces,
                                    sformat='%s DONE %s', symbol='=')))


def get_gender_from_label(gender):
    """
    Method to get the passed gender in XNAT format

    :param gender: gender selected
    :return: value accepted by xnat: 'female' or 'male'
    """
    if gender.lower() in ['female', 'f']:
        return 'female'
    elif gender.lower() in ['male', 'm']:
        return 'male'
    else:
        return 'unknown'


def get_handedness_from_label(handedness):
    """
    Method to get the passed handedness in XNAT format

    :param handedness: handedness selected
    :return: value accepted by xnat: 'right', 'left', or 'ambidextrous'
    """
    if handedness.lower() in ['right', 'r']:
        return 'right'
    elif handedness.lower() in ['left', 'l']:
        return 'left'
    elif handedness.lower() in ['ambidextrous', 'a']:
        return 'ambidextrous'
    else:
        return 'unknown'


def get_resources_list(object_dict, resources_list):
    """
    Method to get the list of resources labels

    :param object_dict: dictionary to describe XNAT object parameters
    :param resources_list: list of resources requested from the user
    :return: None if empty, 'all' if all selected, list otherwise
    """
    # Get list of resources' label
    if resources_list == 'all':
        res_list = object_dict('resources', None)
    else:
        res_list = resources_list
    return res_list


def display_item(project, subject=None, session=None):
    """
    Method to display the tree

    :param project: project ID on XNAT
    :param subject: subject label on XNAT
    :param session: session label on XNAT
    :return: None
    """
    print(('Project: %s' % (project)))
    if subject:
        print(('  +Subject: %s' % (subject)))
        if session:
            print(('    *Session: %s' % (session)))


def is_assessor_type(obj_type):
    """
    Function to return if type is assessor.

    :param obj_dict: dictionary to describe XNAT object parameters
    :return: boolean
    """
    _okeys = list(obj_type.keys())
    return 'xsiType' in _okeys or 'procstatus' in _okeys


def get_obj_info(ind, nb, obj):
    """
    Return info on the object for tools display (upload/download).

    :param ind: index of item to display in the list
    :param nb: number of item in the list of objects
    :param obj: object dictionary in the list
    :return: none
    """
    _format = '  + %d/%d objects: %s'
    _f_scan = 'scan %s %s'
    _f_s_extra = '-- type: %s -- series description: %s -- quality: %s'
    _f_assr = 'proc %s %s '
    _f_a_extra = '-- job status: %s -- QC status: %s'
    if is_assessor_type(obj):
        extra = ''
        _status = obj.get('procstatus', '')
        _qcstatus = obj.get('qcstatus', '')
        if _status or _qcstatus:
            extra = _f_a_extra % (_status, _qcstatus)

        info = _f_assr % (obj['label'], extra)
    else:
        extra = ''
        _type = obj.get('type', '')
        _desc = obj.get('series_description', '')
        _qual = obj.get('quality', '')
        if _type or _desc or _qual:
            extra = _f_s_extra % (_type, _desc, _qual)

        info = _f_scan % (obj['ID'], extra)

    return _format % (ind, nb, info)


def new_tree_object(previous, obj):
    """
    Check that we are still on the same project/subject/session.

    :param previous: previous loop info
    :param obj: object
    :return: True if new info, False otherwise
    """
    if obj['project_id'] != previous['project'] or \
       obj['subject_label'] != previous['subject'] or \
       obj['session_label'] != previous['session']:
        return True

    return False



from unittest import TestCase

import json

from dax import XnatUtils
from dax import assessor_utils

from dax.tests import unit_test_entity_common as common


class XnatUtilsUnitTest(TestCase):

    def test_get_proctype(self):
        proctype, version = XnatUtils.get_proctype(common.spider_tiv_from_gif)
        self.assertEqual(proctype, 'BrainTivFromGIF_v1')
        self.assertEqual(version, '1.0.0')

    def test_get_assessor_inputs(self):
        class TestAssessor:
            class TestAttrs:

                def __init__(self, datatype, property):
                    self.datatype = datatype
                    self.property = property

                def get(self, name):
                    if name == self.datatype + '/' + self.property:
                        return json.dumps({'a': 'b'})
                    else:
                        raise IndexError("it's an index error")

            def __init__(self, datatype, property):
                self.attrs = TestAssessor.TestAttrs(datatype, property)
                self.datatype_ = datatype

            def datatype(self):
                return self.datatype_

        assr = TestAssessor('proc:genProcData', 'inputs')
        assr2 = TestAssessor('something', 'else')
        self.assertEqual(XnatUtils.get_assessor_inputs(assr), {'a': 'b'})
        self.assertEqual(XnatUtils.get_assessor_inputs(assr2), None)

    def test_create_full_assessor_name(self):
        test_entries = [
            ['proj1', 'subj1', 'sess1', '01234567-89ab-cdef-0123-456789abcdef']
        ]

        test_names = ['-x-'.join(t) for t in test_entries]

        for t in range(len(test_entries)):
            name = assessor_utils.full_label(*test_entries[t])
            self.assertEqual(test_names[t], name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" File containing functions called by dax executables """

from datetime import datetime
import imp
import os
import sys
from multiprocessing import Pool
import logging

from . import lockfiles
from . import dax_tools_utils
from . import launcher
from . import assessor_utils
from . import log
from . import XnatUtils
from . import utilities
from . import processors
from .dax_settings import DAX_Settings
from .errors import DaxError
DAX_SETTINGS = DAX_Settings()


def set_logger(logfile, debug):
    """
    Set the logging depth

    :param logfile: File to log output to
    :param debug: Should debug depth be used?
    :return: logger object

    """
    # Logger for logs
    if debug:
        logger = log.setup_debug_logger('dax', logfile)
    else:
        logger = log.setup_info_logger('dax', logfile)
    return logger


def read_settings(settings_path, logger, exe):
    logger.info('Current Process ID: %s' % str(os.getpid()))
    msg = 'Current Process Name: dax.bin.{}({})'
    logger.info(msg.format(exe, settings_path))

    # Load the settings file
    logger.info('loading settings from: %s' % settings_path)
    if settings_path.endswith('.py'):
        settings = imp.load_source('settings', settings_path)
        dax_launcher = settings.myLauncher
    elif settings_path.endswith('.yaml'):
        dax_launcher = read_yaml_settings(settings_path, logger)
    else:
        raise DaxError('Wrong type of settings file given. Please use a \
python file describing the Launcher object or a YAML file.')

    # Run the updates
    logger.info('running launcher, Start Time: %s' % str(datetime.now()))
    return dax_launcher


def launch_jobs(settings_path, logfile, debug, projects=None, sessions=None,
                writeonly=False, pbsdir=None, force_no_qsub=False):
    """
    Method to launch jobs on the grid

    :param settings_path: Path to the project settings file
    :param logfile: Full file of the file used to log to
    :param debug: Should debug mode be used
    :param projects: Project(s) that need to be launched
    :param sessions: Session(s) that need to be updated
    :param writeonly:  write the job files without submitting them
    :param pbsdir: folder to store the pbs file
    :param force_no_qsub: run the job locally on the computer (serial mode)
    :return: None

    """
    # Logger for logs
    logger = set_logger(logfile, debug)

    _launcher_obj = read_settings(settings_path, logger, exe='launch_jobs')
    lockfile_prefix = os.path.splitext(os.path.basename(settings_path))[0]
    try:
        _launcher_obj.launch_jobs(lockfile_prefix, projects, sessions,
                                  writeonly, pbsdir,
                                  force_no_qsub=force_no_qsub)
    except KeyboardInterrupt:
        logger.warn('Killed by user.')
        flagfile = os.path.join(
            os.path.join(_launcher_obj.resdir, 'FlagFiles'),
            '%s_%s' % (lockfile_prefix, launcher.LAUNCH_SUFFIX))
        lockfiles.unlock_flagfile(flagfile)
    except Exception as e:
        logger.critical('Caught exception launching jobs in bin.launch_jobs')
        logger.critical('Exception Class %s with message %s' % (e.__class__,
                                                                str(e)))
        flagfile = os.path.join(
            os.path.join(_launcher_obj.resdir, 'FlagFiles'),
            '%s_%s' % (lockfile_prefix, launcher.LAUNCH_SUFFIX))
        lockfiles.unlock_flagfile(flagfile)

    logger.info('finished launcher, End Time: %s' % str(datetime.now()))


def build(settings_path, logfile, debug, projects=None, sessions=None,
          mod_delta=None, proj_lastrun=None, start_sess=None):
    """
    Method that is responsible for running all modules and putting assessors
     into the database

    :param settings_path: Path to the project settings file
    :param logfile: Full file of the file used to log to
    :param debug: Should debug mode be used
    :param projects: Project(s) that need to be built
    :param sessions: Session(s) that need to be built
    :return: None

    """
    # Logger for logs
    logger = set_logger(logfile, debug)

    _launcher_obj = read_settings(settings_path, logger, exe='build')
    lockfile_prefix = os.path.splitext(os.path.basename(settings_path))[0]
    try:
        _launcher_obj.build(lockfile_prefix, projects, sessions,
                            mod_delta=mod_delta, proj_lastrun=proj_lastrun,
                            start_sess=start_sess)
    except KeyboardInterrupt:
        logger.warn('Killed by user.')
        # TODO: pass in the flag file path
        flagfile = os.path.join(
            os.path.join(_launcher_obj.resdir, 'FlagFiles'),
            '%s_%s' % (lockfile_prefix, launcher.BUILD_SUFFIX))
        lockfiles.unlock_flagfile(flagfile)
    except Exception as e:
        logger.critical('Caught exception building Project in bin.build')
        logger.critical('Exception Class %s with message %s' % (e.__class__,
                                                                str(e)))
        flagfile = os.path.join(
            os.path.join(_launcher_obj.resdir, 'FlagFiles'),
            '%s_%s' % (lockfile_prefix, launcher.BUILD_SUFFIX))
        lockfiles.unlock_flagfile(flagfile)

    logger.info('finished build, End Time: %s' % str(datetime.now()))


def update_tasks(settings_path, logfile, debug, projects=None, sessions=None):
    """
    Method that is responsible for updating a Task.

    :param settings_path: Path to the project settings file
    :param logfile: Full file of the file used to log to
    :param debug: Should debug mode be used
    :param projects: Project(s) that need to be launched
    :param sessions: Session(s) that need to be updated
    :return: None

    """
    # Logger for logs
    logger = set_logger(logfile, debug)

    _launcher_obj = read_settings(settings_path, logger, exe='update_tasks')
    lockfile_prefix = os.path.splitext(os.path.basename(settings_path))[0]
    try:
        _launcher_obj.update_tasks(lockfile_prefix, projects, sessions)
    except KeyboardInterrupt:
        logger.warn('Killed by user.')
        flagfile = os.path.join(
            os.path.join(_launcher_obj.resdir, 'FlagFiles'),
            '%s_%s' % (lockfile_prefix, launcher.UPDATE_SUFFIX))
        lockfiles.unlock_flagfile(flagfile)
    except Exception as e:
        logger.critical('Caught exception updating tasks in bin.update_tasks')
        logger.critical('Exception Class %s with message %s' % (e.__class__,
                                                                str(e)))
        flagfile = os.path.join(
            os.path.join(_launcher_obj.resdir, 'FlagFiles'),
            '%s_%s' % (lockfile_prefix, launcher.UPDATE_SUFFIX))
        lockfiles.unlock_flagfile(flagfile)

    logger.info('finished updating tasks, End Time: %s' % str(datetime.now()))


def read_yaml_settings(yaml_file, logger):
    """
    Method to read the settings yaml file and generate the launcher object.

    :param yaml_file: path to yaml file defining the settings
    :return: launcher object
    """
    if not os.path.isfile(yaml_file):
        err = 'Path not found for {}'
        raise DaxError(err.format(yaml_file))

    doc = utilities.read_yaml(yaml_file)

    # Set Inputs from Yaml
    check_default_keys(yaml_file, doc)

    # Set attributes for settings
    attrs = doc.get('attrs')

    # Set singularity image dir
    singularity_imagedir = doc.get('singularity_imagedir')
    job_template = doc.get('jobtemplate')

    # Read modules
    modulelib = doc.get('modulelib')
    mods = dict()
    modules = doc.get('modules', list())
    for mod_dict in modules:
        if mod_dict.get('filepath') is None:
            err = 'Filepath not set for {}'.format(mod_dict.get('name'))
            raise DaxError(err)

        mod_path = mod_dict.get('filepath')
        if not os.path.isabs(mod_path) and modulelib:
            # Preprend lib location
            mod_path = os.path.join(modulelib, mod_path)

        mods[mod_dict.get('name')] = load_from_file(
            mod_path, mod_dict.get('arguments'), logger)

    # Read processors
    procs = dict()
    processors = doc.get('processors', list())
    for proc_dict in processors:
        if proc_dict.get('filepath') is None:
            err = 'Filepath not set for {}'.format(proc_dict.get('name'))
            raise DaxError(err)
        procs[proc_dict.get('name')] = load_from_file(
            proc_dict.get('filepath'), proc_dict.get('arguments'), logger)

    # Read yaml processors
    processorlib = doc.get('processorlib')
    yamlprocs = dict()
    yamls = doc.get('yamlprocessors', list())
    for yaml_dict in yamls:
        if yaml_dict.get('filepath') is None:
            err = 'Filepath not set for {}'.format(yaml_dict.get('name'))
            raise DaxError(err)

        yaml_path = yaml_dict.get('filepath')
        if not os.path.isabs(yaml_path) and processorlib:
            # Preprend lib location
            yaml_path = os.path.join(processorlib, yaml_path)

        yamlprocs[yaml_dict.get('name')] = load_from_file(
            yaml_path, yaml_dict.get('arguments'),
            logger, singularity_imagedir, job_template)

    # Read sgp processors
    processorlib = doc.get('processorlib')
    sgpprocs = {}
    for s in doc.get('sgpprocessors', []):
        yaml_name = s['name']
        try:
            yaml_path = s['filepath']
        except KeyError:
            raise DaxError('Filepath not set for {}'.format(yaml_name))

        if not os.path.isabs(yaml_path) and processorlib:
            # Preprend lib location
            yaml_path = os.path.join(processorlib, yaml_path)

        yaml_args = s.get('arguments', None)
        sgpprocs[yaml_name] = load_from_file(
            yaml_path,
            yaml_args,
            logger,
            singularity_imagedir,
            job_template)

    # Read projects
    proj_mod = dict()
    proj_proc = dict()
    yaml_proc = dict()
    sgp_processors = {}
    projects = doc.get('projects')
    for proj_dict in projects:
        project = str(proj_dict.get('project'))
        if project:
            # modules:
            if proj_dict.get('modules'):
                for mod_n in proj_dict.get('modules').split(','):
                    if project not in list(proj_mod.keys()):
                        proj_mod[project] = [mods[mod_n]]
                    else:
                        proj_mod[project].append(mods[mod_n])

            # processors:
            if proj_dict.get('processors'):
                for proc_n in [s.strip() for s in proj_dict.get(
                        'processors').split(',')]:
                    if project not in list(proj_proc.keys()):
                        proj_proc[project] = [procs[proc_n]]
                    else:
                        proj_proc[project].append(procs[proc_n])

            # yaml_proc:
            if proj_dict.get('yamlprocessors'):
                for yaml_n in [s.strip() for s in proj_dict.get(
                        'yamlprocessors').split(',')]:
                    if project not in list(yaml_proc.keys()):
                        yaml_proc[project] = [yamlprocs[yaml_n]]
                    else:
                        yaml_proc[project].append(yamlprocs[yaml_n])

            # sgp processors:
            _list = proj_dict.get('sgpprocessors', '').split(',')
            _list = [s.strip() for s in _list if s]
            for yaml_n in _list:
                if project not in list(sgp_processors.keys()):
                    sgp_processors[project] = [sgpprocs[yaml_n]]
                else:
                    sgp_processors[project].append(sgpprocs[yaml_n])

    # set in attrs:
    attrs['project_process_dict'] = proj_proc
    attrs['project_modules_dict'] = proj_mod
    attrs['yaml_dict'] = yaml_proc
    attrs['project_sgp_processors'] = sgp_processors

    # Delete unsupported arguments
    attrs.pop('skip_lastupdate', None)

    attrs['job_template'] = doc.get('jobtemplate')
    attrs['timeout_emails'] = doc.get('timeout_emails')
    attrs['smtp_host'] = doc.get('smtp_host')

    # Return a launcher with specified arguments
    resdir = doc.get('resdir')
    return launcher.Launcher(resdir, **attrs)


def check_default_keys(yaml_file, doc):
    """ Static method to raise error if key not found in dictionary from
    yaml file.
    :param yaml_file: path to yaml file defining the processor
    :param doc: doc dictionary extracted from the yaml file
    """
    # processors / modules / yamlprocessors are not required
    for key in ['projects', 'attrs']:
        raise_yaml_error_if_no_key(doc, yaml_file, key)


def raise_yaml_error_if_no_key(doc, yaml_file, key):
    """Method to raise an execption if the key is not in the dict
    :param doc: dict to check
    :param yaml_file: YAMLfile path
    :param key: key to search
    """
    if key not in list(doc.keys()):
        err = 'YAML File {} does not have {} defined. See example.'
        raise DaxError(err.format(yaml_file, key))


def load_from_file(filepath, args, logger, singularity_imagedir=None, job_template=None):
    """
    Check if a file exists and if it's a python file
    :param filepath: path to the file to test
    :return: True the file pass the test, False otherwise
    """
    if args is not None:
        _tmp = 'test.{}(**args)'
    else:
        _tmp = 'test.{}()'

    if not os.path.isfile(filepath):
        raise DaxError('File %s does not exists.' % filepath)

    if filepath.endswith('.py'):
        test = imp.load_source('test', filepath)
        try:
            #print('evaling:' + _tmp.format(os.path.basename(filepath)[:-3]))
            return eval(_tmp.format(os.path.basename(filepath)[:-3]))
        except AttributeError as e:
            #print('attribute error', str(e))
            pass

        err = '[ERROR] Module or processor NOT FOUND in the python file {}.'
        logger.error(err.format(filepath))

    elif filepath.endswith('.yaml'):
        return processors.load_from_yaml(
            XnatUtils, filepath, args, singularity_imagedir, job_template)

    return None


def upload(settings_path, max_threads=1):
    logger = logging.getLogger('dax')

    # Load settings from file
    _launcher_obj = read_yaml_settings(settings_path, logger)

    host = _launcher_obj.xnat_host
    resdir = _launcher_obj.resdir

    lock_dir = os.path.join(resdir, 'FlagFiles')

    # TODO: filter based on projects/sessions

    # Load list of assessors to be uploaded
    upload_queue = launcher.load_task_queue(resdir, status='COMPLETE')
    upload_queue += launcher.load_task_queue(resdir, status='JOB_FAILED')

    alist = [_.assessor_label for _ in upload_queue]
    acount = len(alist)

    # Clean lock files
    lockfiles.clean_lockfiles(lock_dir, logger)

    # Count running uploads
    lock_list = os.listdir(lock_dir)
    lock_list = [x for x in lock_list if x.endswith('_Upload.txt')]
    cur_upload_count = len(lock_list)
    logger.info('count of already running uploads:' + str(cur_upload_count))

    num_threads = max_threads - cur_upload_count
    if num_threads < 1:
        logger.info('max uploads already:{}'.format(str(cur_upload_count)))
        return

    logger.info('starting {} upload thread(s)'.format(str(num_threads)))
    sys.stdout.flush()

    pool = Pool(processes=num_threads)
    for aindex, alabel in enumerate(alist):
        sys.stdout.flush()
        pool.apply_async(upload_thread, [host, aindex, alabel, acount, resdir])

    logger.info('waiting for upload pool to finish...')
    sys.stdout.flush()

    pool.close()
    pool.join()

    logger.info('upload pool finished')
    sys.stdout.flush()


def upload_thread(xnat_host, pindex, assessor_label, pcount, resdir):
    # TODO: move this and associated functions to launcher
    logger = logging.getLogger('dax')

    lock_dir = os.path.join(resdir, 'FlagFiles')

    # Get lock file name
    lock_file = os.path.join(lock_dir, '{}_Upload.txt'.format(assessor_label))

    # Try to lock
    success = lockfiles.lock_flagfile(lock_file)
    if not success:
        # Failed to get lock
        logger.warn('failed to get lock:{}'.format(lock_file))
        return

    try:
        logger.info('connecting to xnat for upload:{}'.format(xnat_host))
        # Upload the assessor
        with XnatUtils.get_interface(xnat_host) as xnat:
            msg = '*Upload Process:{}/{}:{}'.format(
                pindex + 1, pcount, assessor_label)
            logger.info(msg)

            assessor_path = os.path.join(resdir, assessor_label)

            if assessor_utils.is_sgp_assessor(assessor_label):
                # It's a subject gen proc assessor, handle it specifically
                dax_tools_utils.upload_assessor_subjgenproc(
                    xnat, assessor_path)
            else:
                assessor_dict = assessor_utils.parse_full_assessor_name(
                    assessor_label)
                if assessor_dict:
                    uploaded = dax_tools_utils.upload_assessor(
                        xnat, assessor_dict, assessor_path, resdir)
                    if not uploaded:
                        msg = 'not uploaded:{}'.format(assessor_label)
                        logger.warn(msg)
                else:
                    logger.warn('     --> wrong label')
    except Exception as err:
        logger.error('error uploading:{}'.format(err))

    finally:
        # Delete the lock file
        logger.debug('deleting lock file:{}'.format(lock_file))
        lockfiles.unlock_flagfile(lock_file)


def undo_processing(assessor_label, logger=None):
    """
    Unset job information for the assessor on XNAT, Delete files, set to run.

    :return: None

    """

    if not logger:
        logger = logging.getLogger()

    logger.info('undo assessor')

    with XnatUtils.get_interface() as xnat:

        # First use the assessor label to find the assessor on XNAT
        if assessor_utils.is_sgp_assessor(assessor_label):
            logger.info('connect to sgp assessor')
            adict = assessor_utils.parse_full_assessor_name(assessor_label)
            assr = xnat.select_sgp_assessor(
                adict['project_id'],
                adict['subject_label'],
                adict['label'])
        else:
            logger.info('connect to genproc assessor')
            adict = assessor_utils.parse_full_assessor_name(assessor_label)
            assr = xnat.select_assessor(
                adict['project_id'],
                adict['subject_label'],
                adict['session_label'],
                adict['label'])

        # Reset job related fields
        logger.info('setting job info to null')
        xsitype = assr.datatype().lower()
        assr.attrs.set('{}/{}'.format(xsitype, 'jobnode'), 'null')
        assr.attrs.set('{}/{}'.format(xsitype, 'jobid'), 'null')
        assr.attrs.set('{}/{}'.format(xsitype, 'jobstartdate'), 'null')
        assr.attrs.set('{}/{}'.format(xsitype, 'memused'), 'null')
        assr.attrs.set('{}/{}'.format(xsitype, 'walltimeused'), 'null')
        assr.attrs.mset({
            '{}/{}'.format(xsitype, 'procstatus'): 'NEED_INPUTS',
            '{}/{}'.format(xsitype, 'validation/status'): 'Job Pending'})

        # Get list of file resources
        if assessor_utils.is_sgp_assessor(assessor_label):
            resources = assr.resources()
        else:
            resources = assr.out_resources()

        # Delete the resources
        for res in resources:
            res_label = res.label()

            if res_label in ['OLD', 'EDITS']:
                logger('skipping resource:{}'.format(res_label))
                continue

            logger.info('removing:{}'.format(res_label))
            try:
                res.delete()
            except Exception:
                logger.error('deleting resource')



""" Task Builder for DAX in REDCap."""

import os
import logging
import re
import fnmatch
import tempfile

from dax.task import NeedInputsException, NoDataException
from dax.task import JOB_PENDING, JOB_RUNNING
from dax.task import NEED_INPUTS, NEED_TO_RUN, NO_DATA
from dax.processors import load_from_yaml, SgpProcessor
from .projectinfo import load_project_info
from .taskqueue import TaskQueue

logger = logging.getLogger('manager.rcq.taskbuilder')


PROCESSING_RENAME = {
    'redcap_repeat_instance': 'ID',
    'processor_file': 'FILE',
    'processor_filter': 'FILTER',
    'processor_args': 'ARGS',
    'processing_complete': 'COMPLETE',
}


class TaskBuilder(object):
    # Builds tasks and adds to queue in REDCap

    def __init__(self, projects_redcap, xnat, yamldir):
        self._projects_rc = projects_redcap
        self._xnat = xnat
        self._yamldir = yamldir
        self._queue = TaskQueue(projects_redcap)
        self._build_projects = self.projects_with_processing()

    def update(self, project):

        if project not in self._build_projects:
            logger.info(f'no processing protocols found:{project}')
            return

        with tempfile.TemporaryDirectory() as tmpdir:

            logger.debug(f'loading processing protools:{project}')
            protocols = self._load_protocols(project, tmpdir)

            if len(protocols) == 0:
                logger.debug(f'no processing protocols found:{project}')
                return

            info = load_project_info(self._xnat, project)

            # Iterate processing protocols
            for i, row in enumerate(protocols):
                filepath = row['FILE']

                logger.info(f'{project}:{filepath}')

                if row.get('ARGS', False):
                    user_inputs = row.get('ARGS')
                    logger.debug(f'overrides:{user_inputs}')
                    rlist = user_inputs.strip().split('\r\n')
                    rdict = {}
                    for arg in rlist:
                        try:
                            key, val = arg.split(':', 1)
                            rdict[key] = val.strip()
                        except ValueError as e:
                            msg = f'invalid arg:{project}:{filepath}:{arg}:{e}'
                            raise Exception(msg)

                    user_inputs = rdict
                    logger.debug(f'user_inputs:{user_inputs}')
                else:
                    user_inputs = None

                if row['FILTER']:
                    include_filters = row['FILTER'].replace(' ', '').split(',')
                else:
                    include_filters = []

                logger.debug(f'building processor:{filepath}')
                self._build_processor(
                    filepath,
                    user_inputs,
                    info,
                    include_filters,
                    custom=row['CUSTOM'])

    def _build_processor(self, filepath, user_inputs, info, include_filters, custom=False):
        # Get lists of subjects/sessions for filtering

        # Load the processor
        logger.debug(f'loading processor from yaml:{filepath}')
        try:
            processor = load_from_yaml(
                self._xnat,
                filepath,
                user_inputs=user_inputs,
                job_template='~/job_template.txt')
        except Exception as err:
            logger.debug(f'failed to load, cannot build:{filepath}:{err}')
            return

        if not processor:
            logger.error(f'loading processor:{filepath}')
            return

        if isinstance(processor, SgpProcessor):
            # Handle subject level processing

            # Get list of subjects to process
            if include_filters:
                include_subjects = _filter_labels(
                    info['all_subjects'], include_filters)
            else:
                include_subjects = info['all_subjects']

            logger.debug(f'include subjects={include_subjects}')

            # Apply the processor to filtered sessions
            for subj in sorted(include_subjects):
                logger.debug(f'subject:{subj}')
                self._build_subject_processor(
                    processor, subj, info, custom=custom)
        else:
            # Handle session level processing

            # Get list of sessions to process
            if include_filters:
                include_sessions = _filter_labels(
                    info['all_sessions'], include_filters)
            else:
                include_sessions = info['all_sessions']

            logger.debug(f'include sessions={include_sessions}')

            # Apply the processor to filtered sessions
            for sess in sorted(include_sessions):
                self._build_session_processor(
                    processor, sess, info, custom=custom)

    def projects_with_processing(self):
        protocols = []
        def_field = self._projects_rc.def_field

        rec = self._projects_rc.export_records(
            forms=['processing'],
            fields=[def_field])

        rec = [x for x in rec if x['redcap_repeat_instrument'] == 'processing']

        # Only enabled processing
        rec = [x for x in rec if str(x['processing_complete']) == '2']

        projects = [x[def_field] for x in rec]

        return projects

    def _load_protocols(self, project, tmpdir):
        protocols = []
        def_field = self._projects_rc.def_field
        yamldir = self._yamldir

        rec = self._projects_rc.export_records(
            records=[project],
            forms=['processing'],
            fields=[def_field])

        rec = [x for x in rec if x['redcap_repeat_instrument'] == 'processing']

        # Only enabled processing
        rec = [x for x in rec if str(x['processing_complete']) == '2']

        for r in rec:
            # Initialize record with project
            d = {'PROJECT': r[def_field]}

            # Find the yaml file
            if r['processor_yamlupload']:
                filepath = r['processor_yamlupload']
                filepath = self._save_processor_file(
                    project,
                    r['redcap_repeat_instance'],
                    tmpdir)
                d['CUSTOM'] = True
            else:
                filepath = r['processor_file']
                d['CUSTOM'] = False

            if not os.path.isabs(filepath):
                # Prepend lib location
                filepath = os.path.join(yamldir, filepath)

            if not os.path.isfile(filepath):
                logger.info(f'file not found:{filepath}')
                continue

            # Get renamed variables
            for k, v in PROCESSING_RENAME.items():
                d[v] = r.get(k, '')

            d['FILE'] = filepath
            d['TYPE'] = _get_proctype(d['FILE'])

            d['EDIT'] = 'edit'

            # Finally, add to our list
            protocols.append(d)

        return protocols

    def _save_processor_file(self, project, repeat_id, outdir):
        # Get the file contents from REDCap
        try:
            (cont, hdr) = self._projects_rc.export_file(
                record=project,
                field='processor_yamlupload',
                repeat_instance=repeat_id)

            if cont == '':
                raise Exception('error exporting file from REDCap')

        except Exception as err:
            logger.error(f'downloading file:{err}')
            return None

        # Save contents to local file
        filename = os.path.join(outdir, hdr['name'])
        try:
            with open(filename, 'wb') as f:
                f.write(cont)

            return filename
        except FileNotFoundError as err:
            logger.error(f'file not found:{filename}:{err}')
            return None

    def _build_session_processor(self, processor, session, project_info, custom=False):
        # Get list of inputs sets (not yet matched with existing)
        inputsets = processor.parse_session_pd(session, project_info)

        logger.debug(f'{session}:{processor.name}')

        logger.debug(inputsets)
        for inputs in inputsets:
            if inputs == {}:
                # Blank inputs
                return

            # Get(create) assessor with given inputs and proc type
            (assr, info) = processor.get_assessor(
                session, inputs, project_info)

            if info['PROCSTATUS'] in [NEED_TO_RUN, NEED_INPUTS]:
                logger.debug('building task')
                (assr, info) = self._build_task(
                    assr, info, processor, project_info, custom=custom)

                logger.debug(f'{info}')
                logger.debug(
                    'status:{}:{}'.format(info['ASSR'], info['PROCSTATUS']))
            else:
                logger.debug('already built:{}'.format(info['ASSR']))

    def _build_subject_processor(self, processor, subject, project_info, custom=False):
        logger.debug(f'{subject}:{processor.name}')
        # Get list of inputs sets (not yet matched with existing)
        inputsets = processor.parse_subject(subject, project_info)
        logger.debug(inputsets)

        for inputs in inputsets:

            if inputs == {}:
                # Blank inputs
                return

            # Get(create) assessor with given inputs and proc type
            (assr, info) = processor.get_assessor(
                self._xnat, subject, inputs, project_info)

            if info['PROCSTATUS'] in [NEED_TO_RUN, NEED_INPUTS]:
                logger.debug('building task')
                (assr, info) = self._build_task(
                    assr, info, processor, project_info, custom=custom)

                logger.debug(f'assr after={info}')
            else:
                logger.debug('already built:{}'.format(info['ASSR']))

    def _build_task(self, assr, info, processor, project_info, custom=False):
        '''Build a task, create assessor in XNAT, add new record to redcap'''
        old_proc_status = info['PROCSTATUS']
        old_qc_status = info['QCSTATUS']

        try:
            var2val, inputlist = processor.build_var2val(
                assr,
                info,
                project_info)

            self._queue._add_task(
                project_info['name'],
                info['ASSR'],
                inputlist,
                var2val,
                processor.walltime_str,
                processor.memreq_mb,
                processor.yaml_file,
                processor.user_inputs,
                custom=custom
            )

            # Set new statuses to be updated
            new_proc_status = JOB_RUNNING
            new_qc_status = JOB_PENDING
        except NeedInputsException as e:
            new_proc_status = NEED_INPUTS
            new_qc_status = e.value
        except NoDataException as e:
            new_proc_status = NO_DATA
            new_qc_status = e.value

        # Update on xnat
        _xsitype = processor.xsitype.lower()
        if new_proc_status != old_proc_status:
            assr.attrs.set(f'{_xsitype}/procstatus', new_proc_status)
        if new_qc_status != old_qc_status:
            assr.attrs.set(f'{_xsitype}/validation/status', new_qc_status)

        # Update local info
        info['PROCSTATUS'] = new_proc_status
        info['QCSTATUS'] = new_qc_status

        return (assr, info)


def _filter_matches(match_input, match_filter):
    return re.match(fnmatch.translate(match_filter), match_input)


def _filter_labels(labels, filters):
    filtered_labels = []

    for f in filters:
        filtered_labels += [x for x in labels if _filter_matches(x, f)]

    return list(set(filtered_labels))


def _get_proctype(procfile):
    # Get just the filename without the directory path
    tmp = os.path.basename(procfile)

    # Split on periods and grab the 4th value from right,
    # thus allowing periods in the main processor name
    return tmp.rsplit('.')[-4]
